from melodine.models.base.track import TrackBase
from melodine.models.base.artist import ArtistBase
from melodine.models.base.album import AlbumBase
from melodine.models.base.playlist import PlaylistBase

__all__ = (
    'TrackBase',
    'ArtistBase',
    'AlbumBase',
    'PlaylistBase'
)



from dataclasses import dataclass, field
from typing import Iterable, List, Optional

from melodine.utils import SearchResults
from melodine.configs import SPOTIFY
from melodine.models.spotify import (
    Album,
    Artist,
    Episode,
    Playlist,
    Show,
    Track
)

__all__ = ["search"]

_TYPES = {"artist": Artist, "album": Album, "track": Track,
          "playlist": Playlist, "show": Show, "episode": Episode}
_SEARCH_TYPES = {"artists", "albums", "tracks", "playlists", "shows", "episodes"}

@dataclass
class SpotifySearchResults(SearchResults):
    ''' A dataclass for Search Results

    Inherits from the base SearchResults class.

    Attributes
    ----------
    artists : List[:class:`Artist`]
        The artists of the search.
    playlists : List[:class:`Playlist`]
        The playlists of the search.
    albums : List[:class:`Album`]
        The albums of the search.
    tracks : List[:class:`Track`]
        The tracks of the search.
    videos: List[:class:`Video`]
        The videos from the search results
    '''
    
    artists: Optional[List[Artist]] = field(default_factory=list)
    albums: Optional[List[Album]] = field(default_factory=list)
    tracks: Optional[List[Track]] = field(default_factory=list)
    playlists: Optional[List[Playlist]] = field(default_factory=list)

    shows: Optional[List[Show]] = field(default_factory=list)
    episodes: Optional[List[Episode]] = field(default_factory=list)


def search(
    q: str,  # pylint: disable=invalid-name
    *,
    types: Iterable[str] = ("tracks", "playlists", "artists", "albums", "shows", "episodes"),
    limit: int = 20,
    offset: int = 0,
) -> SpotifySearchResults:
    '''Get search results for a query'''

    if types is None:
        types = ("track", "playlists", "artists", "albums", "shows", "episodes")

    if not hasattr(types, "__iter__"):
        raise TypeError("types must be an iterable.")

    types_ = set(types)

    if not types_.issubset(_SEARCH_TYPES):
        raise ValueError('Bad queary type! got "%s" expected any of: tracks, playlists, artists, albums, shows, episodes' %
                         types_.difference(_SEARCH_TYPES).pop())

    types = list(map(lambda type: type[:-1] if type.endswith('s') else type, types))
    
    query_type = ",".join(tp.strip() for tp in types)

    data = SPOTIFY.search(
        q=q,
        limit=limit,
        offset=offset,
        type=query_type,
    )

    return SpotifySearchResults(
        **{
            key: [_TYPES[_val['type']](_val) for _val in value['items']]
            for key, value in data.items()
        }
    )



from dataclasses import dataclass
from datetime import datetime
from typing import List

from melodine.utils import Image, URIBase
from melodine.configs import SPOTIFY


@dataclass
class Episode(URIBase):
    '''A spotify Show (Podcast) Episode object'''

    def __init__(self, data: dict) -> None:
        from .show import Show

        self.id: str = data.get('id', str())  # pylint: disable=invalid-name
        self.href: str = f'https://open.spotify.com/episode/{self.id}'
        self.uri: str = f'spotify:episode:{self.id}'

        self.name: str = data.get('name', str())
        self.description: str = data.get('description', str())
        self.duration: int = data.get('duration', int())
        self.release_data: datetime = datetime.strptime(
            data.get('release_date', '1000-10-10'), "%Y-%m-%d")

        self._show = None if data.get('show') is None else Show(data.get('show'))
        self.images: List[Image] = [
            Image(**image) for image in data.get('images', [])
        ]

    def __repr__(self) -> str:
        return f'<spotify.Episode: {self.name!r}>'

    @property
    def show(self):
        from .show import Show

        if self._show is not None:
            return self.show

        self._show = Show(SPOTIFY.episode(self.id)['show'])
        return self._show

    def add_to_playlist(self, playlist_id) -> None:
        SPOTIFY.playlist_add_items(
            playlist_id=playlist_id,
            items=[self.id]
        )

    # def is_saved(self) -> bool:
    #     return SPOTIFY.current_user_saved_episodes_contains([self.id])[0]

    # def save_episode(self) -> None:
    #     return SPOTIFY.current_user_saved_episodes_add([self.id])

    # def unsave_episode(self) -> None:
    #     return SPOTIFY.current_user_saved_episodes_delete([self.id])



from melodine.models.spotify.album import Album
from melodine.models.spotify.artist import Artist
from melodine.models.spotify.track import Track
from melodine.models.spotify.playlist import Playlist
from melodine.models.spotify.show import Show
from melodine.models.spotify.episode import Episode
from melodine.models.spotify.device import Device
from melodine.models.spotify.user import User
from melodine.models.spotify.player import Player
from melodine.models.spotify._search import search
from melodine.models.spotify.client import client

# client = Client()

__all__ = [
    'Track',
    'Artist',
    'Album',
    'Playlist',
    'User',
    'Player',
    'Show',
    'Episode',
    'Device',
    'search',
    'client'
]



import datetime
from typing import List, Optional

from typing_extensions import Self

from melodine.configs import SPOTIFY, YTMUSIC
from melodine.innertube import InnerTube
from melodine.models.spotify.artist import Artist
from melodine.models.spotify.episode import Episode
from melodine.utils import Image, URIBase
from melodine.models.base.track import TrackBase


class Track(URIBase):
    '''A Spotify Track Object

    Attributes
    ---------
    id: str
        The Spotify ID for the Track
    '''

    __slots__ = [
        "artists",
        "album",
        "id",
        "name",
        "uri",
        "href",
        "duration",
        "url_",
        "images",
        "explicit"
    ]

    def __new__(cls: Self, data, **kwargs):
        # for some cases when a result contains an episode
        if data.get('episode'):
            return Episode(data)
        return super().__new__(cls)

    def __init__(self, data, **kwargs) -> None:
        from melodine.models.spotify.album import Album

        self.artists = list(
            Artist(_artist) for _artist in data.get('artists', [])
        )

        self.album = Album(
            data.get('album', {})
        ) if 'album' in data else kwargs.get('album', {})

        self.id = data.get('id', None)  # pylint: disable=invalid-name
        self.name = data.get('name', None)
        self.uri = data.get('uri', None)
        self.href = 'https://open.spotify.com/track/' + self.id
        self.duration = int(data.get('duration_ms') / 1000)
        self.explicit = data.get('explicit', False)
        self.url_ = []

        if 'images' in data:
            self.images = [
                Image(**image) for image in data.get('images', [])
            ]
        else:
            self.images = self.album.images.copy()

        self._audio_analysis = {}
        self._audio_features = {}

    def __repr__(self) -> str:
        return f"<melo.Track - {(self.name or self.id or self.uri)!r}>"

    def test_run():  # pylint: disable=no-method-argument
        return Track(SPOTIFY.search('ritchrd - paris', type='track')['tracks']['items'][0])

    @property
    def url(self):
        '''porperty getter for the Track URL'''

        if self.url_:
            return self.url_

        video_id = YTMUSIC.search(
            f"{self.artists[0].name} - {self.name}", filter='songs')[0]['videoId']

        # for ID of the top search results which could be a video (that's not implemented yet)
        # video_id = YTMUSIC.search(f"{self.artists[0].name} - {self.name}")[0]['videoId']

        video_info = InnerTube().player(video_id)

        # self.url_ = video_info['streamingData']['adaptiveFormats'][-1]['url']
        self.url_ = video_info['streamingData']['formats'][-1]['url']

        return self.url_

    def cache_url(self) -> None:
        '''just a dummy call to trigger the url fetching'''
        if not self.url:
            self.url = self.url

    def get_recommendations(
        self,
        limit: Optional[int] = 1
    ) -> List["Track"]:
        ''' get recommendation for a particular track

        Returns one Track per call, use the limit parameter to change number of returned tracks.

        Parameters
        ----------
        limit: int
            The number of suggestions to returns

        Returns
        -------
        results: List[Tracks]
            A list of suggested tracks
        '''
        recs = SPOTIFY.recommendations(seed_tracks=[self.id], limit=limit)

        return [Track(rec) for rec in recs['tracks']]

    def audio_analysis(self):
        '''get audio analysis for a track based on spotify community's listening patterns'''
        if self._audio_analysis:
            return self._audio_analysis

        self._audio_analysis = SPOTIFY.audio_analysis(self.id)
        return self._audio_analysis

    def audio_features(self):
        '''get audio features for a track based on spotify community's listening patterns'''
        if self._audio_features:
            return self._audio_features

        self._audio_analysis = SPOTIFY.audio_features(self.id)
        return self.audio_features

    def add_to_playlist(self, playlist_id) -> None:
        '''Add the track to a spotify playlist'''
        SPOTIFY.playlist_add_items(
            playlist_id=playlist_id,
            items=[self.id]
        )

        # def is_saved(self) -> bool:
        #     return SPOTIFY.current_user_saved_tracks_contains([self.id])[0]

        # def save_track(self) -> None:
        #     return SPOTIFY.current_user_saved_tracks_add([self.id])

        # def unsave_track(self) -> None:
        #     return SPOTIFY.current_user_saved_tracks_delete([self.id])


class PlaylistTrack(Track):
    '''A playlist track.

    same as a normal track, but with some extra attributes
    '''
    __slots__ = ['added_by', 'added_at']

    def __init__(self, data, **kwargs) -> None:
        from melodine.models.spotify.client import Client
        from melodine.models.spotify.user import User

        super().__init__(data['track'])

        self.added_by = User(data.get('added_by')) if not isinstance(
            data.get('added_by'), (Client, User)) else data.get('added_by')
        self.added_at = datetime.datetime.strptime(
            data["added_at"], "%Y-%m-%dT%H:%M:%SZ"
        )

    def __repr__(self):
        return f"<spotify.PlaylistTrack: {self.name!r}>"



from typing import Dict, List, Union

from melodine.configs import YTMUSIC
from melodine.models import ytmusic
from melodine.utils import Image, URIBase
from melodine.models.base.artist import ArtistBase


class Artist(URIBase):

    __slots__ = (
        '_data',
        'id',
        '_name',
        'href',
        'uri',
        '_shuffle_id',
        '_radio_id',
        '_singles',
        '_videos',
        '_albums',
        '_songs',
        '_params',
        '_images',
        '_related_artists',
        '_followers',
        '_description',
        '_views',
    )

    def __init__(self, data: dict, **kwargs) -> None:
        self._data = None

        self.id: str = data.get('channelId')
        self._name: Union[str, None] = data.get('name', None)
        self.href: str = 'https://music.youtube.com/channel/' + \
            str(self.id or '')
        self.uri: str = 'ytmusic:artist:' + str(self.id or '')

        self._shuffle_id: str = data.get('shuffleId', None)
        self._radio_id: str = data.get('radioId', None)

        self._singles: List[Dict] = data.get('singles', {}).get('results', [])
        self._videos: List[Dict] = data.get('videos', {}).get('results', [])
        self._albums: List[Dict] = data.get('albums', {}).get('results', [])
        self._songs: List[Dict] = data.get('songs', {}).get('results', [])

        self._params: str = data.get('singles', {}).get('params', None)

        self._images: List[Image] = [
            Image(**image)
            for image in data.get('images', [])
        ]

        self._related_artists: List[Dict] = data.get(
            'related', {}).get('results', [])
        self._followers: Union[str, None] = data.get('subscribers', None)
        self._description: Union[str, None] = data.get('description', None)
        self._views: Union[str, None] = data.get('views', None)

        self.total_albums = len(self._albums) + len(self._singles)

    def _get_data(self):
        self._data = YTMUSIC.get_artist(self.id)

    @classmethod
    def from_id(cls, id: str):
        return cls(data={'channelId': id})

    @classmethod
    def partial(cls, data: Dict) -> "Artist":
        '''
        for artist data obtained from a track or an album.
            `{'name': '', 'id': ''}`
        '''
        return cls(data={
            'name': data['name'],
            'channelId': data['id']
        })

    @classmethod
    def from_search(cls, data: Dict) -> "Artist":
        '''
        for artist data obtained from a search result 
            `{'artist': '', 'browseId': '', 'shuffleId': '', 'radioId': ''}`

        also accomodates for artist data from artist['related']
            `{'title': '', 'browseId': '', 'subscribers': '', 'thumbnails': ''}`
        '''
        data['name'] = data.pop('artist', data.pop('title', str()))
        data['channelId'] = data.pop('browseId')
        return cls(data=data)

    @property
    def name(self):
        if self._name is None:
            if self._data is None:
                self._get_data()
            self._name = self._data['name']
        return self._name

    @property
    def shuffle_id(self) -> str:
        if self._shuffle_id is None:
            if self._data is None:
                self._get_data()
            self._shuffle_id = self._data['shuffleId']
        return self._shuffle_id

    @property
    def radio_id(self) -> str:
        if self._radio_id is None:
            if self._data is None:
                self._get_data()
            self._radio_id = self._data['radioId']
        return self._radio_id

    @property
    def singles(self) -> List["ytmusic.Album"]:
        '''
        singles and albums are essentially the same thing, 
        hence comparing them against albums
        '''
        from .album import Album

        if len(self._singles) == 0:
            if self._data is None:
                self._get_data()
            self._singles = self._data['singles']['results']
        for idx, single in enumerate(self._singles.copy()):
            if not isinstance(single, Album):
                single = Album(single)
                self._singles.insert(idx, single)
                del self._singles[idx + 1]
        return self._singles

    @property
    def videos(self) -> List["ytmusic.Video"]:
        from .video import Video

        if len(self._videos) == 0:
            if self._data is None:
                self._get_data()
            self._videos = self._data['videos']['results']
        for idx, video in enumerate(self._videos.copy()):
            if not isinstance(video, Video):
                video = Video(video)
                self._videos.insert(idx, video)
                del self._videos[idx + 1]
        return self._videos

    @property
    def albums(self) -> List["ytmusic.Album"]:
        from .album import Album

        if len(self._albums) == 0:
            if self._data is None:
                self._get_data()
            self._albums = self._data['albums']['results']
        for idx, album in enumerate(self._albums.copy()):
            if not isinstance(album, Album):
                album = Album(album)
                self._albums.insert(idx, album)
                del self._albums[idx + 1]
        return self._albums

    @property
    def tracks(self) -> List["ytmusic.Track"]:
        from .track import Track

        if len(self._songs) == 0:
            if self._data is None:
                self._get_data()
            self._songs = self._data['songs']['results']
        for idx, track in enumerate(self._songs.copy()):
            if not isinstance(track, Track):
                track = Track(track)
                self._songs.insert(idx, track)
                del self._songs[idx + 1]
        return self._songs

    @property
    def params(self):
        if self._params is None:
            if self._data is None:
                self._get_data()
            self._params = self._data['singles']['params']
        return self._params

    @property
    def images(self) -> List[Image]:
        if len(self._images) == 0:
            if self._data is None:
                self._get_data()
            self._images = [
                Image(**image)
                for image in self._data['thumbnails']
            ]
        return self._images

    @property
    def related_artists(self) -> List["ytmusic.Artist"]:
        if len(self._related_artists) == 0:
            if self._data is None:
                self._get_data()
            self._related_artists = self._data['related']['results']
        for idx, artist in enumerate(self._related_artists.copy()):
            if not isinstance(artist, Artist):
                artist = Artist.from_search(artist)
                self._related_artists.insert(idx, artist)
                del self._related_artists[idx + 1]
        return self._related_artists

    @property
    def followers(self) -> str:
        if self._followers is None:
            if self._data is None:
                self._get_data()
            self._followers = self._data['subscribers']
        return self._followers

    @property
    def description(self) -> str:
        if self._description is None:
            if self._data is None:
                self._get_data()
            self._desctiption = self._data.get('description')
        return self._desctiption

    @property
    def views(self) -> int:
        if self._views is None:
            if self._data is None:
                self._get_data()
            self._views = int(
                self._data['views']
                .split(' ')[0]
                .replace(',', '_')
            )
        return self._views



import json
import urllib.parse
from typing import List, Literal

import requests

# Extracted API keys -- unclear what these are linked to.
_api_keys: List[str] = [
    'AIzaSyAO_FJ2SlqU8Q4STEHLGCilw_Y9_11qcW8',
    'AIzaSyCtkvNIR1HCEwzsqK6JuE6KqpyjusIRI30',
    'AIzaSyA8eiZmM1FaDVjRy-df2KTyQ_vz_yYM39w',
    'AIzaSyC8UYZpvA2eknNex0Pjid0_eTLJoDu6los',
    'AIzaSyCjc_pVEDi4qsv5MtC2dMXzpIaDoRFLsxw',
    'AIzaSyDHQ9ipnphqTzDqZsbtd8_Ru4_kiKVQe2k'
]

_default_clients = {
    'WEB': {
        'context': {
            'client': {
                'clientName': 'WEB',
                'clientVersion': '2.20200720.00.02'
            }
        },
        'api_key': 'AIzaSyAO_FJ2SlqU8Q4STEHLGCilw_Y9_11qcW8'
    },
    'ANDROID': {
        'context': {
            'client': {
                'clientName': 'ANDROID',
                'clientVersion': '16.20'
            }
        },
        'api_key': 'AIzaSyAO_FJ2SlqU8Q4STEHLGCilw_Y9_11qcW8'
    },
    'WEB_EMBED': {
        'context': {
            'client': {
                'clientName': 'WEB',
                'clientVersion': '2.20210721.00.00',
                'clientScreen': 'EMBED'
            }
        },
        'api_key': 'AIzaSyAO_FJ2SlqU8Q4STEHLGCilw_Y9_11qcW8'
    },
    'ANDROID_EMBED': {
        'context': {
            'client': {
                'clientName': 'ANDROID',
                'clientVersion': '16.20',
                'clientScreen': 'EMBED'
            }
        },
        'api_key': 'AIzaSyAO_FJ2SlqU8Q4STEHLGCilw_Y9_11qcW8'
    }
}


class InnerTube:
    '''A class to interact with the InnerTube API'''

    def __init__(
        self,
        client: Literal['ANDROID', 'WEB',
                        'ANDROID_EMBED', 'WEB_EMBED'] = 'ANDROID'
    ) -> None:
        self.context = _default_clients[client]['context']
        self.api_key = _default_clients[client]['api_key']

    @property
    def base_url(self) -> str:
        '''returns the base url endpoint for the InnerTube API'''
        return 'https://www.youtube.com/youtubei/v1'

    @property
    def base_data(self) -> str:
        '''returns the base json data to send to the InnerTube API'''
        return {
            'context': self.context
        }

    @property
    def base_params(self) -> str:
        '''Return the base query parameters to transmit to the innertube API.'''
        return {
            'key': self.api_key,
            'contentCheckOk': True,
            'racyCheckOk': True
        }

    def _call_api(
        self,
        endpoint: str,
        query: dict,
        data: dict
    ) -> dict:
        '''Make a request to the given endpoint with the query and data'''
        endpoint_url = f'{endpoint}?{urllib.parse.urlencode(query)}'
        headers = {
            "Content-Type": "application/json",
            "User-Agent": "Mozilla/5.0",
            "accept-language": "en-US,en",
        }

        data = bytes(json.dumps(data), encoding="utf-8")

        response = requests.post(
            endpoint_url,
            headers=headers,
            data=data
        )

        return json.loads(response.text)

    def player(
        self,
        video_id: str
    ) -> dict:
        '''Get details for a Video.

        Parameters
        ----------
        video_id : `str`
            The id of the video to get the player info for.

        Returns
        -------
        response : `dict`
            raw player info json.
        '''
        endpoint = f'{self.base_url}/player'
        query = {
            'videoId': video_id,
        }
        query.update(self.base_params)

        return self._call_api(endpoint, query, self.base_data)

    def search(
        self,
        query: str,
        continuation: bool = None
    ) -> dict:
        '''Search for a track based on a query
        
        Parameters
        ----------
        query: str
            The query to search for.
        continuation: bool
            Weather to autocomplete queries based on suggestions.

        Returns
        -------
        results: List[Tracks]
            A list of search results
        '''
        endpoint = f'{self.base_url}/search'
        query = {
            "query": query,
        }
        query.update(self.base_params)
        if continuation:
            self.base_data['continuation'] = continuation
        
        return self._call_api(endpoint, query, self.base_data)

    def recommendations(
        self,
        video_id: str,
    ) -> dict:
        '''Get recommendations for a tracks complementary to a playlist

        Parameters
        ----------
        video_id : str
            ID of the video to get recommendations for.

        playlist_id : str
            ID of the playlist to get the recommendations relative to

        Returns
        -------
        response : dict
            The response form the API containig the track suggestions
        '''
        endpoint = f'{self.base_url}/music/get_queue'
        query = {
            'videoId': video_id,
            'playlistId': f'RDAMVM{video_id}',
        }
        query.update(self.base_params)

        return self._call_api(endpoint, query, self.base_data)


if __name__ == '__main__':
    tube = InnerTube()

    # player = tube.player('ZJ2RydH1EbE')
    # streams = player['streamingData']['adaptiveFormats'][-1]
    # url = streams['url']



from dataclasses import dataclass
from functools import cached_property
from typing import Dict, List, Optional, Union

from melodine.utils import Image, URIBase
from melodine.configs import SPOTIFY
from melodine.models.spotify.episode import Episode


@dataclass
class Show(URIBase):
    '''A spotify Show (Podcast) object'''
    def __init__(self, data: Union[Dict, str]) -> None:
        self.id: str = data.get('id', str())  #pylint: disable=invalid-name
        self.href: str = f'https://open.spotify.com/show/{self.id}'
        self.uri: str = f'spotify:show:{self.id}'

        self.name: str = data.get('name', str())
        self.description: str = data.get('description', str())
        self.total_episodes: int = data.get('total_episodes', int())
        self.publisher: str = data.get('published', str())

        self.images: List[Image] = [
            Image(**image) for image in data.get('images', [])
        ]

    def __repr__(self) -> str:
        return f'<spotify.Show: {self.name!r}>'

    @property
    def episodes(
        self,
        *,
        limit: Optional[int] = 10,
        offset: Optional[int] = 0
    ) -> List[Episode]:
        '''Get paged results for episodes from a show based on the limit and offset'''

        data = SPOTIFY.show_episodes(
            self.id,
            limit=limit,
            offset=offset
        )

        return [Episode(episode) for episode in data['items']]

    @cached_property
    def get_all_episodes(self) -> List[Episode]:
        '''Get a list of all the episodes of a show'''
        result = []
        offset = 0

        while len(result) < self.total_episodes:
            data = SPOTIFY.show_episodes(
                self.id,
                limit=10,
                offset=offset
            )
            result += [Episode(episode) for episode in data['items']]
            offset += 10

        return result

    # def is_saved(self) -> bool:
    #     return SPOTIFY.current_user_saved_shows_contains([self.id])[0]

    # def save_show(self) -> None:
    #     return SPOTIFY.current_user_saved_shows_add([self.id])

    # def unsave_show(self) -> None:
    #     return SPOTIFY.current_user_saved_shows_delete([self.id])



from typing import List, Optional

from melodine.utils import Image, URIBase
from melodine.configs import SPOTIFY
from melodine.models.spotify.artist import Artist
from melodine.models.spotify.track import Track
from melodine.models.base.album import AlbumBase


class Album(URIBase):
    '''
    id - spotify id for the album
    name - the name of the album
    artists - a list containing a artist object of all the artists from a album
    tracks - a list containing track objects for all the tracks in a album
    type - the type of album (a full album, or a single or smthng)
    '''

    __slots__ = (
        "_data",
        "_tracks",
        "artists",
        "images",
        "href",
        "name",
        "type",
        "data",
        'year',
        "uri",
        "id",
    )

    def __init__(self, data):

        self._data = data
        self.id = data.get('id', None)  # pylint: disable=invalid-name
        self.name = data.get('name')
        self.href = data.get('external_urls').get('spotify', None)
        self.uri = data.get('uri', None)

        self.year = data.get('release_date').split('-')[0]
        self.type = data.get('album_type', None)
        self.images = [Image(**image) for image in data.get('images', [])]

        self.artists = [Artist(artist) for artist in data.get('artists')]
        self._track_count = data.get('total_tracks', 0)
        self._tracks = []

    def __repr__(self) -> str:
        return f"melo.Album - {(self.name or self.id or self.uri)!r}"

    @property
    def total_tracks(self) -> int:
        '''get all the tracks from an album'''

        if not self._total_tracks:
            self._total_tracks = SPOTIFY.album_tracks(self.id, limit=1)[
                'total']
        return self._total_tracks

    @property
    def tracks(self):
        '''get all the tracks from the album'''
        offset = 0

        while len(self._tracks) < self.total_tracks:
            data = SPOTIFY.album_tracks(self.id, limit=50, offset=offset)
            offset += 50
            self._tracks += list(Track(track, album=self._data)
                                 for track in data['items'])
        return self._tracks

    def get_tracks(self, limit: Optional[int] = 20, offset: Optional[int] = 0) -> List[Track]:
        '''get specific tracks from an album based on the limit and offsets'''
        if len(self._tracks) == 0:
            data = SPOTIFY.album_tracks(self.id, limit=limit, offset=offset)
            for track in data['items']:
                track['album'] = self._data
            self._tracks = list(Track(_track) for _track in data['items'])
        return self._tracks

    # @cached_property
    # def get_all_tracks(self) -> List[Track]:
    #     '''get all tracks of an album'''
    #     offset = 0

    #     while len(self.tracks) < self.total_tracks:
    #         data = SPOTIFY.album_tracks(self.id, limit=50, offset=offset)
    #         offset += 50
    #         self.tracks += list(Track(track, album=self._data) for track in data['items'])
    #     return self.tracks

    # def saved(self) -> bool:
    #     ''' Check if one or more albums is already saved in
    #         the current Spotify user’s “Your Music” library.
    #     '''
    #     return SPOTIFY.current_user_saved_albums_contains([self.id])[0]

    # def save_album(self) -> None:
    #     '''Add one or more albums to the current user's
    #         "Your Music" library.
    #     '''
    #     return SPOTIFY.current_user_saved_albums_add([self.id])

    # def unsave_album(self):
    #     return SPOTIFY.current_user_saved_albums_delete([self.id])



from dataclasses import dataclass, field
from itertools import groupby
from typing import Iterable, List, Literal, Optional

from melodine.configs import YTMUSIC
from melodine.models.ytmusic import Album, Artist, Playlist, Track, Video
from melodine.utils import SearchResults

__all__ = ['search']

_TYPES = {"artist": Artist, "album": Album,
          "song": Track, "video": Video, "playlist": Playlist}
_SEARCH_TYPES = {"artists", "albums", "tracks", "videos", "playlists"}


@dataclass
class YTMusicSearchResults(SearchResults):
    ''' A dataclass for Search Results

    Inherits from the SearchResultsBase class for.

    Attributes
    ----------
    artists : List[:class:`Artist`]
        The artists of the search.
    playlists : List[:class:`Playlist`]
        The playlists of the search.
    albums : List[:class:`Album`]
        The albums of the search.
    tracks : List[:class:`Track`]
        The tracks of the search.
    videos: List[:class:`Video`]
        The videos from the search results
    '''

    tracks: Optional[List[Track]] = field(default_factory=list)
    albums: Optional[List[Album]] = field(default_factory=list)
    artists: Optional[List[Artist]] = field(default_factory=list)
    playlists: Optional[List[Playlist]] = field(default_factory=list)

    videos: Optional[List[Video]] = field(default_factory=list)


def search(
    q: str,
    *,
    types: Optional[Iterable[Literal['tracks', 'videos',
                                     'albums', 'artists', 'playlists']]] = [],
    limit: Optional[int] = 15,
) -> YTMusicSearchResults:

    data = []

    if types:

        if not hasattr(types, "__iter__"):
            raise TypeError("types must be an iterable.")

        types = set(types)

        if not types.issubset(_SEARCH_TYPES):
            raise ValueError('Bad queary type! got "%s" expected one or more of: tracks, playlists, artists, albums, videos' %
                             types.difference(_SEARCH_TYPES).pop())

        if 'tracks' in types:
            types.remove('tracks')
            types.add('songs')

        for type in types:
            results = YTMUSIC.search(
                q,
                filter=type,
                limit=limit
            )
            data.extend(results)
    else:
        results = YTMUSIC.search(q)
        data.extend(results)

    search_results = dict.fromkeys(_SEARCH_TYPES, [])
    def key_func(x): return x['resultType']
    for key, value in groupby(sorted(data, key=key_func), key_func):
        search_results[key + 's'] = list(value)

    search_results['tracks'] = search_results.pop('songs', [])

    return YTMusicSearchResults(
        **{
            ('tracks' if key == 'song' else key if key.endswith('s') else key + 's'): [
                _TYPES[_val['resultType']](_val)
                if _val['resultType'] != 'artist'
                else _TYPES[_val['resultType']].from_search(_val)
                for _val in value
            ]
            for key, value in search_results.items()
        }
    )



from melodine.models.ytmusic.track import Track
from melodine.models.ytmusic.artist import Artist
from melodine.models.ytmusic.album import Album
from melodine.models.ytmusic.video import Video
from melodine.models.ytmusic.playlist import Playlist
from melodine.models.ytmusic._search import search
from melodine.models.ytmusic.user import User

__all__ = ['Track', 'Artist', 'Album', 'Video', 'User', 'Playlist', 'search']



from abc import ABCMeta


class PlaylistBase(metaclass=ABCMeta):
    @classmethod
    def __subclasshook__(cls, subclass):
        return (
            hasattr(subclass, 'id') and
            hasattr(subclass, 'name') and
            hasattr(subclass, 'href') and
            hasattr(subclass, 'uri') and
            hasattr(subclass, 'owner') and
            hasattr(subclass, 'description') and
            hasattr(subclass, 'images') and

            hasattr(subclass, 'total_tracks') and
            hasattr(subclass, 'tracks')
        )

    # @property
    # @abstractmethod
    # def id(self):
    #     raise NotImplementedError

    # @property
    # @abstractmethod
    # def name(self):
    #     raise NotImplementedError

    # @property
    # @abstractmethod
    # def href(self):
    #     raise NotImplementedError

    # @property
    # @abstractmethod
    # def uri(self):
    #     raise NotImplementedError

    # @property
    # @abstractmethod
    # def owner(self):
    #     raise NotImplementedError

    # @property
    # @abstractmethod
    # def description(self):
    #     raise NotImplementedError

    # @property
    # @abstractmethod
    # def total_tracks(self):
    #     raise NotImplementedError

    # @property
    # @abstractmethod
    # def tracks(self):
    #     raise NotImplementedError



from dataclasses import dataclass
import functools


class Image:
    """
    An object representing a Spotify image resource.

    Attributes
    ----------
    height : :class:`str`
        The height of the image.
    width : :class:`str`
        The width of the image.
    url : :class:`str`
        The URL of the image.
    """

    __slots__ = ("height", "width", "url")

    def __init__(self, *, height: str, width: str, url: str):
        self.height = height
        self.width = width
        self.url = url

    def __repr__(self):
        return f"<melo.Image: {self.url!r} (width: {self.width!r}, height: {self.height!r})>"

    def __eq__(self, other):
        return type(self) is type(other) and self.url == other.url


@dataclass
class SearchResults:
    """ A base dataclass of the YTMusicSearchResults and SpotifySearchResults classes. """

    def __add__(self, other: "SearchResults"):
        self_items = self.__dict__.items()
        for attr, val in other.__dict__.items():
            if attr not in self_items:
                self.__setattr__(attr, val)
            else:
                self.__setattr__(attr, self.__getattribute__(attr).extend(val))
        return self

    def __bool__(self):
        return any(self.__dict__.values())

    def __repr__(self) -> str:
        return f"<melo.SearchResults: {id(self)}>"


class URIBase:
    '''
    A base class to define basic dunder methods for all URI baed methods. It's purpose is to reduce boiler plate for models

    All melo models must inherit from this class to have basic dunder methods
    '''

    uri = repr(None)
    id = repr(None)

    def __hash__(self):
        return hash(self.uri)

    def __eq__(self, __o: object) -> bool:
        return (
            type(self) is type(__o) and self.uri == __o.uri
        )

    def __ne__(self, __o: object) -> bool:
        return not self.__eq__(__o)

    def __bool__(self):
        return not not self.id

    def __str__(self) -> str:
        return self.uri


def singleton(cls):
    '''class definitions marked singleton remember created instances and return that one same instance each time its object is instantialted.

    i.e. only one instance of an object can exist at any given time.'''
    @functools.wraps(cls)
    def wrapper_singleton(*args, **kwargs):
        if not wrapper_singleton.instance:
            wrapper_singleton.instance = cls(*args, **kwargs)
        return wrapper_singleton.instance
    wrapper_singleton.instance = None
    return wrapper_singleton



"""
a playlist is owned only by users.

some artists may own playlists but they dont show up the artists's profile.
so a simple try catch for when  a artist owns a playlist shoudl suffice.
"""

from typing import Dict, List, Union

from melodine.configs import YTMUSIC
from melodine.models import ytmusic
from melodine.models.ytmusic.artist import Artist
from melodine.models.ytmusic.track import Track
from melodine.models.ytmusic.video import Video
from melodine.utils import Image, URIBase
from melodine.models.base.playlist import PlaylistBase


class Playlist(URIBase):

    __slots__ = (
        '_data',
        'id',
        '_name',
        'href',
        'uri',
        '_tracks',
        '_owner',
        '_duration',
        '_total_tracks',
        '_images',
        '_description',
        '_year',
    )

    def __init__(self, data: Dict):
        self._data = None

        self.id: str = data.get('id', data.get(
            'playlistId', data.get('browseId')))  # pylint: disable=invalid-name
        if self.id.startswith('VL'):
            self.id = self.id[2:]
        self._name: str = data.get('title', None)
        self.href: str = "https://music.youtube.com/playlist?list=" + self.id
        self.uri: str = "ytmusic:user:" + self.id

        self._tracks: List[Dict] = data.get('tracks', [])
        self._owner: Union[Dict, str] = data.get('author', [])
        self._duration: Union[int, None] = data.get('duration_seconds', None)
        self._total_tracks: Union[int, None] = data.get(
            'trackCount', data.get('itemCount'))

        self.images: List[Image] = [
            Image(**image)
            for image in data.get('thumbnails', [])
        ]

        self._description = data.get('description', None)
        self._year = data.get('year', None)

    def _get_data(self) -> None:
        self._data = YTMUSIC.get_playlist(self.id)

    @classmethod
    def from_id(cls, id: str) -> "Playlist":
        return cls(data={'id': id})

    @property
    def name(self) -> str:
        if self._name is None:
            if self._data is None:
                self._get_data()
            self._name = self._data['title']
        return self._name

    @property
    def tracks(self) -> List[Union[Video, Track]]:
        if len(self._tracks) == 0:
            if self._data is None:
                self._get_data()
            self._tracks = self._data['tracks']
        for idx, video in enumerate(self._tracks.copy()):
            if not isinstance(video, (Video, Track)):
                if video['videoId'] is not None:
                    if video.get('album') is not None:
                        video = Track(video)
                    else:
                        video = Video(video)
                    self._tracks.insert(idx, video)
                    del self._tracks[idx + 1]
                else:
                    self._tracks[idx] = None
        self._tracks = list(
            filter(lambda track: track is not None, self._tracks)
        )
        return self._tracks

    @property
    def owner(self) -> Union["ytmusic.User", Artist]:
        from .user import User

        if isinstance(self._owner, str) or not self._owner:
            if self._data is None:
                self._get_data()
            self._owner = self._data['author']

        if not isinstance(self._owner, (Artist, User)):
            try:
                user = YTMUSIC.get_user(self._owner if isinstance(
                    self._owner, str) else self._owner['id'])
                self._owner = User.from_id(self._owner) if isinstance(
                    self._owner, str) else User(user, id=self._owner['id'])
            except (KeyError, ValueError, Exception):
                self._owner = Artist.from_id(self._owner) if isinstance(
                    self._owner, str) else Artist(self._owner)
        return self._owner

    @property
    def duration(self) -> int:
        if self._duration is None:
            if self._data is None:
                self._get_data()
            self._duration = self._data['duration_seconds']
        return self._duration

    @property
    def total_tracks(self) -> int:
        if self._total_tracks is None:
            if self._data is None:
                self._get_data()
            self._total_tracks = self._data['trackCount']
        return self._total_tracks

    @property
    def description(self) -> str:
        if self._description is None:
            if self._data is None:
                self._get_data()
            self._description = self._data['description']
        return str(self._description or '')

    @property
    def year(self) -> int:
        if self._year is None:
            if self._data is None:
                self._get_data()
            self._year = int(self._data['year'])
        return self._year



"""
only playlists and some videos have users.

videos might belong to either artists or users. so a try-catch here too.
albeit, the try-catch wrapped api call needs to be only for the data attr (only when it's needed).
"""

from typing import Dict, List

from melodine.utils import URIBase
from melodine.configs import YTMUSIC, YT
from melodine.models import ytmusic
from melodine.models.ytmusic import Video


class User(URIBase):

    __slots__ = (
        '_data',
        'id',
        '_name',
        'href',
        'uri',
        '_videos',
        '_playlists',

        '_total_playlists',
        '_total_uploads',
        '__next_page_token',
        '__prev_page_token'
    )

    def __init__(self, data: Dict, **kwargs) -> None:
        self._data = None

        self.id: str = data.get('id', kwargs.get('id', None))
        self.uploads_id: str = 'UU' + self.id[2:]
        self._name: str = data.get('name', None)
        self.href: str = "https://music.youtube.com/channel/" + \
            str(self.id or '')
        self.uri: str = "ytmusic:user:" + str(self.id or '')

        self._videos: List[Dict] = data.get('videos', {}).get('results')
        self._playlists: List[Dict] = data.get('playlists', {}).get('results')

        self._total_playlists = int()  # aka 0
        self._total_uploads = int()
        self.__next_page_token = None
        self.__prev_page_token = None

    def _get_data(self) -> None:
        self._data = YTMUSIC.get_user(self.id)

    @classmethod
    def from_id(cls, data: Dict) -> "User":
        return cls(data={'id': data})

    @property
    def total_uploads(self) -> int:
        '''total number of track uploaded by the user'''
        if self._total_uploads:
            return self._total_uploads

        self._total_uploads = YT.get_playlist_items(
            playlist_id=self.uploads_id,
            limit=1,
            return_json=True
        )['pageInfo']['totalResults']
        return self._total_uploads

    @property
    def total_playlists(self) -> int:
        '''total number of playlists from the user'''
        if self._total_playlists:
            return self._total_playlists

        self._total_playlists = YT.get_playlists(
            channel_id=self.id,
            limit=1,
            return_json=True
        )['pageInfo']['totalResults']
        return self._total_playlists

    @property
    def name(self) -> str:
        if self._name is None:
            if self._data is None:
                self._get_data()
            self._name = self._data['name']
        return self._name

    @property
    def videos(self) -> List[Video]:
        if len(self._videos) == 0:
            if self._data is None:
                self._get_data()
            self._videos = self._data['videos']['results']
        for idx, video in enumerate(self._videos.copy()):
            if not isinstance(video, Video):
                video = Video(video)
                self._videos.insert(idx, video)
                del self._videos[idx + 1]
        return self._videos

    def get_videos(
        self,
        limit: int = 10,
        reset=False
    ) -> List[Video]:
        '''Get a list of Videos from a User'''
        results = []
        data = {
            'nextPageToken': (
                '<placeholder>'
                if not self.__prev_page_token else self.__next_page_token
            )
        }

        if data.get('nextPageToken'):
            if data['nextPageToken'] == '<placeholder>' or reset:
                data = YT.get_playlist_items(
                    playlist_id=self.uploads_id,
                    limit=limit,
                    return_json=True
                )
            else:
                data = YT.get_playlist_items(
                    playlist_id=self.uploads_id,
                    page_token=data['nextPagetoken'],
                    limit=limit,
                    return_json=True
                )

            self.__next_page_token = data['nextPageToken']
            self.__prev_page_token = data['prevPageToken']

            _results = [Video(video['snippet']) for video in data['items']]
            results += _results
            return _results

    def get_all_videos(self) -> List[Video]:
        '''gets a list of all videos uploaded by a user.

        Might take extremely long time based on the number of videos an artist has.
        (some channels end up having over 100 uploads)

        Use :func:`get_videos` for paged results instead
        '''
        if len(self._videos) == self.total_uploads:
            return self.videos

        count = 0
        data = {}

        self._videos = []
        while count < self.total_uploads:
            if data:
                data = YT.get_playlist_items(
                    playlist_id=self.uploads_id, page_token=data['nextPageToken'], return_json=True)
            else:
                data = YT.get_playlist_items(
                    playlist_id=self.uploads_id, return_json=True)
            self._videos += [
                Video(video['snippet'])
                for video in data['items']
            ]
            count += len(data['items'])
        return self._videos

    @property
    def playlists(self) -> List["ytmusic.Playlist"]:
        from .playlist import Playlist

        if len(self._playlists) == 0:
            if self._data is None:
                self._get_data()
            self._playlists = self._data['playlists']['results']
        for idx, playlist in enumerate(self._playlists.copy()):
            if not isinstance(playlist, Playlist):
                playlist = Playlist(playlist)
                self._playlists.insert(idx, playlist)
                del self._playlists[idx + 1]
        return self._playlists

    def get_all_playlists(self) -> List["ytmusic.Playlist"]:
        '''Get a list of all Playlists from a User'''
        from .playlist import Playlist

        data = {'nextPageToken': '<placeholder>'}

        self._playlists = []
        while data.get('nextPageToken'):
            if not data:
                data = YT.get_playlists(
                    channel_id=self.id,
                    limit=5,
                    return_json=True
                )
            else:
                data = YT.get_playlists(
                    channel_id=self.id,
                    limit=5,
                    page_token=data['nextPageToken'],
                    return_json=True
                )

            self._playlists += [
                Playlist(playlist['id'])
                for playlist in data['items']
            ]
        return self._playlists



from dataclasses import dataclass
from typing import Dict, List

from melodine.models import spotify  # pylint: disable=unused-import

from melodine.models.spotify.device import Device
from melodine.models.spotify.track import Track


@dataclass
class PlayerContext:
    '''the "Playing From" context for particular playback state'''

    def __init__(self, data: Dict) -> None:
        self.external_urls: List[Dict[str, str]] = data.get('external_urls')
        self.type: str = data.get('type')
        self.uri: str = data.get('uri')
        self.href: str = f'open.spotify.com/{type}/{self.uri.split(":")[2]}'


class Player:
    def __init__(self, data: Dict, user: "spotify.User") -> None:
        self.user = user

        self.device: Device = Device(data.get('device'))
        self.shuffle_state: bool = data.get('shuffle_state')
        self.repeat_state: bool = data.get('repeat_state')
        self.context: PlayerContext = data.get('context')
        self.progress = data.get('progress_ms') * 1000
        self.item: Track = Track(data.get('item'))
        self.currently_playing_type = data.get('currently_playing_type')
        self.is_playing: bool = data.get('is_playing')

    def __repr__(self):
        return f"<spotify.Player: {self.user!r}>"

    # TODO - check if basic player modifications works for non-premium users and implement those methods



from typing import Any, List, Union

from melodine.models.ytmusic import Video
from melodine.configs import YT


def search(
    q: str,  #pylint: disable=invalid-name
    *,
    related_to_video_id: Union[str, None] = None,
    limit: int = 10,
) -> List[Video]:

    result: List[Union[Video, Any]] = []

    data = YT.search(
        q=q,
        related_to_video_id=related_to_video_id,
        limit=limit,
        return_json=True
    )

    for video in data['items']:
        video['snippet']['videoId'] = video['id']['videoId']
        if 'thumbnails' in video['snippet']:
            video['snippet']['thumbnails'] = list(video['snippet']['thumbnails'].values())
        video = Video(video['snippet'])
        result.append(video)

    return result



import os
from dataclasses import dataclass
from functools import cached_property
from typing import Dict, List, Literal, Optional

import appdirs
import spotipy
from typing_extensions import Self

from melodine.configs import CACHE_PATH, SCOPES, SPOTIFY
from melodine.models.spotify.album import Album
from melodine.models.spotify.artist import Artist
from melodine.models.spotify.episode import Episode
from melodine.models.spotify.player import Player
from melodine.models.spotify.playlist import Playlist
from melodine.models.spotify.show import Show
from melodine.models.spotify.track import PlaylistTrack, Track
from melodine.utils import Image, URIBase, singleton


@singleton
@dataclass
class Category:
    def __init__(self, data: Dict) -> None:
        self.id = data.get('id')  # pylint: disable=invalid-name
        self.name = data.get('name')
        self.images = [Image(**image) for image in data.get('icons')]

    @classmethod
    def from_id(cls, category_id: str) -> Self:
        category_data = SPOTIFY.category(category_id)
        return cls(category_data)

    def get_playlist(self) -> List[Playlist]:
        data = SPOTIFY.category_playlists(self.id)
        return [Playlist(playlist) for playlist in data['playlists']['items']]


class Client(URIBase):
    def __init__(self) -> None:
        if not os.path.exists(CACHE_PATH):
            self.__is_authorized = False

            self.id = None  # pylint: disable=invalid-name
            self.name = None
            self.href = None
            self.uri = None
            self.followers = None
            self.type = None
            self.images = None
        else:
            self.authorize()

    def __repr__(self):
        return f"<spotify.Client: {(self.name or 'Unauthorized')!r}>"

    @property
    def is_authorized(self) -> bool:
        return self.__is_authorized

    def authorize(self) -> None:
        '''carry out authorization to access a user's data'''

        SPOTIFY = spotipy.Spotify(
            auth_manager=spotipy.SpotifyOAuth(
                scope=SCOPES,
                client_id="22e27810dff0451bb93a71beb5e4b70d",
                client_secret="6254b7703d8540a48b4795d82eae9300",
                redirect_uri="http://localhost:8080/",
                cache_handler=spotipy.CacheFileHandler(
                    cache_path=os.path.join(
                        appdirs.user_data_dir(), '.melo', 'spotify_cache'
                    )
                )
            )
        )

        data = SPOTIFY.current_user()

        self.id = data.get('id')
        self.name = data.get('display_name')
        self.href = 'open.spotify.com/user/' + self.id
        self.uri = data.get('uri')
        self.followers = data.get('followers')
        self.type = data.get('type')
        self.images = [Image(**image_) for image_ in data.get('images', [])]
        self.__is_authorized = True

    def currently_playing(self) -> Track:
        data = SPOTIFY.current_user_playing_track()
        return Track(data['item'])

    def current_playback(self) -> Player:
        return Player(SPOTIFY.current_playback(), self)

    def recently_played(self, limit: int = 20) -> List[Track]:
        data = SPOTIFY.current_user_recently_played(limit=limit)
        return [Track(track_['track']) for track_ in data.get('items')]

    def featured_playlist(self) -> List[Playlist]:
        data = SPOTIFY.featured_playlists()
        return [Playlist(playlist) for playlist in data['playlists']['items']]

    def new_releases(self) -> List[Album]:
        data = SPOTIFY.new_releases()
        return [Album(album) for album in data['albums']['items']]

    @cached_property
    def categories(self):
        data = SPOTIFY.categories()
        return [Category(category) for category in data['categories']['items']]
    
    def is_artist_followed(self, artist_id: str) -> bool:
        return SPOTIFY.current_user_following_artists([artist_id])

    def follow_artist(self, artist_id: str) -> None:
        return SPOTIFY.user_follow_artists([artist_id])

    def unfollow_artist(self, artist_id: str) -> None:
        return SPOTIFY.user_unfollow_artists([artist_id])

    def followed_artists(
        self,
        *,
        limit: Optional[int] = 20,
    ) -> List[Artist]:
        data = SPOTIFY.current_user_followed_artists(limit=limit)
        return [Artist(artist) for artist in data['artists']['items']]

    @cached_property
    def all_followed_artists(self) -> List[Artist]:
        data = {'next': '<placeholder>'}
        results = []

        while data.get('next'):
            if data['next'] == '<placeholder>':
                data = SPOTIFY.current_user_followed_artists()['artists']
            else:
                data = SPOTIFY.next(data)['artists']
            results += data['items']
        return [Artist(artist) for artist in results]
    
    def is_playlist_followed(self, playlist_id: str) -> bool:
        return SPOTIFY.playlist_is_following(playlist_id, [self.id])

    def save_playlist(self, playlist_id: str) -> None:
        '''Add the current user as the follower of this playlist'''
        SPOTIFY.current_user_follow_playlist(playlist_id)

    def unsave_playlist(playlist_id: str) -> None:
        return SPOTIFY.current_user_unfollow_playlist(playlist_id)

    def saved_playlists(
        self,
        *,
        limit: Optional[int] = 50,
        offset: Optional[int] = 0
    ) -> List[Playlist]:
        data = SPOTIFY.current_user_playlists(limit=limit, offset=offset)
        return [Playlist(playlist) for playlist in data['items']]

    @cached_property
    def all_saved_playlists(self) -> List[Playlist]:
        data = {'next': '<placeholder>'}
        results = []

        while data.get('next'):
            if data['next'] == '<placeholder>':
                data = SPOTIFY.current_user_playlists()
            else:
                data = SPOTIFY.next(data)
            results += data['items']
        return [Playlist(playlist) for playlist in results]


    def is_album_saved(self, album_id: str) -> bool:
        ''' Check if an album is already saved in
            the current Spotify user’s “Your Music” library.
        '''
        return SPOTIFY.current_user_saved_albums_contains([album_id])[0]

    def save_album(self, album_id: str) -> None:
        '''Add one or more albums to the current user's
            "Your Music" library.
        '''
        return SPOTIFY.current_user_saved_albums_add([album_id])

    def unsave_album(self, album_id: str):
        return SPOTIFY.current_user_saved_albums_delete([album_id])
    
    def saved_albums(
        self,
        *,
        limit: Optional[int] = 20,
        offset: Optional[int] = 0
    ) -> List[Album]:
        data = SPOTIFY.current_user_saved_albums(limit=limit, offset=offset)
        return [Album(album['album']) for album in data['items']]

    @cached_property
    def all_saved_albums(self) -> List[Album]:
        data = {'next': '<placeholder>'}
        results = []

        while data.get('next'):
            if data['next'] == '<placeholder>':
                print('lmao')
                data = SPOTIFY.current_user_saved_albums()
                print(data.keys())
            else:
                data = SPOTIFY.next(data)
            results += data['items']
        return [Album(album['album']) for album in results]
    
    def is_track_saved(self, track_id: str) -> bool:
        return SPOTIFY.current_user_saved_tracks_contains([track_id])[0]

    def save_track(self, track_id: str) -> None:
        return SPOTIFY.current_user_saved_tracks_add([track_id])

    def unsave_track(self, track_id: str) -> None:
        return SPOTIFY.current_user_saved_tracks_delete([track_id])

    def saved_tracks(
        self,
        *,
        limit: Optional[int] = 20,
        offset: Optional[int] = 0
    ) -> List[PlaylistTrack]:
        data = SPOTIFY.current_user_saved_tracks(limit=limit, offset=offset)
        results = []
        for track in data['items']:
            track['added_by'] = self
            track = PlaylistTrack(track)
            results.append(track)
        return results

    @cached_property
    def all_saved_tracks(self) -> List[PlaylistTrack]:
        data = {'next': '<placeholder>'}
        results = []

        while data.get('next'):
            if data['next'] == '<placeholder>':
                data = SPOTIFY.current_user_saved_tracks()
            else:
                data = SPOTIFY.next(data)
            results += data['items']
        for idx, track in enumerate(results.copy()):
            track['added_by'] = self
            track = PlaylistTrack(track)
            results[idx] = track
        return results
    
    def is_episode_saved(self, episode_id: str) -> bool:
        return SPOTIFY.current_user_saved_episodes_contains([episode_id])[0]

    def save_episode(self, episode_id: str) -> None:
        return SPOTIFY.current_user_saved_episodes_add([episode_id])

    def unsave_episode(self, episode_id: str) -> None:
        return SPOTIFY.current_user_saved_episodes_delete([episode_id])

    def saved_episodes(
        self,
        *,
        limit: Optional[int] = 20,
        offset: Optional[int] = 0
    ) -> List[Episode]:
        data = SPOTIFY.current_user_saved_episodes(limit=limit, offset=offset)
        return [Episode(episode['episode']) for episode in data['items']]

    def all_saved_episodes(self) -> List[Episode]:
        data = {'next': '<placeholder>'}
        results = []

        while data.get('next'):
            if data['next'] == '<placeholder>':
                data = SPOTIFY.current_user_saved_episodes()
            else:
                data = SPOTIFY.next(data)
            results += data['items']
        return [Episode(episode['episode']) for episode in results]

    def is_show_saved(self, show_id: str) -> bool:
        return SPOTIFY.current_user_saved_shows_contains([show_id])[0]

    def save_show(self, show_id: str) -> None:
        return SPOTIFY.current_user_saved_shows_add([show_id])

    def unsave_show(self, show_id: str) -> None:
        return SPOTIFY.current_user_saved_shows_delete([show_id])

    def saved_shows(
        self,
        *,
        limit: Optional[int] = 20,
        offset: Optional[int] = 0
    ) -> List[Show]:
        data = SPOTIFY.current_user_saved_shows(limit=limit, offset=offset)
        return [Show(show['show']) for show in data['items']]

    def all_saved_shows(self) -> List[Show]:
        data = {'next': '<placeholder>'}
        results = []

        while data.get('next'):
            if data['next'] == '<placeholder>':
                data = SPOTIFY.current_user_saved_shows()
            else:
                data = SPOTIFY.next(data)
            results += data['items']
        return [Show(show['show']) for show in results]

    def top_artists(
        self,
        limit: Optional[int] = 20,
        offset: Optional[int] = 0,
        time_range: Literal['long_term', 'short_term',
                            'medium_term'] = 'medium_term'
    ) -> List[Artist]:

        data = SPOTIFY.current_user_top_artists(
            limit=limit,
            offset=offset,
            time_range=time_range
        )

        return [Artist(artist_) for artist_ in data['items']]

    def top_tracks(
        self,
        limit: Optional[int] = 20,
        offset: Optional[int] = 0,
        time_range: Literal[
            'long_term','short_term', 'medium_term'
        ] = 'medium_term'
    ) -> List[Track]:

        data = SPOTIFY.current_user_top_tracks(
            limit=limit,
            offset=offset,
            time_range=time_range
        )

        return [Track(track_) for track_ in data['items']]


client = Client()



from functools import cached_property
from typing import Dict, List, Optional

from melodine.configs import SPOTIFY
from melodine.models.spotify.playlist import Playlist
from melodine.utils import Image, URIBase


class User(URIBase):
    def __init__(self, data: Dict) -> None:
        self.id = data.get('id')  #pylint: disable=invalid-name
        self.name = data.get('display_name')
        self.href = 'open.spotify.com/user/' + self.id
        self.uri = data.get('uri')
        self.followers = data.get('followers')
        self.type = data.get('type')
        self._images = [Image(**image_) for image_ in data.get('images', [])]

    def __repr__(self):
        return f"<spotify.User: {(self.name or self.id)!r}>"

    @property
    def images(self):
        '''thumbnails for a user's profile'''
        if self._images:
            return self._images

        data = SPOTIFY.user(self.id)
        self._images = [Image(**image) for image in data.get('images', [])]
        return self._images

    def user_playlists(
        self,
        limit: Optional[int] = 30,
        offset: Optional[int] = 0
    ) -> List[Playlist]:
        '''Get a user's playlists'''
        data = SPOTIFY.user_playlists(
            self.id,
            limit=limit,
            offset=offset
        )

        return [Playlist(playlist) for playlist in data.get('items', [])]

    @cached_property
    def all_user_playlists(self) -> List[Playlist]:
        '''Get all playlists from a user'''
        data = {'next': '<placeholder>'}
        results = []

        while data.get('next'):
            if data['next'] == '<placeholder>':
                data = SPOTIFY.user_playlists(self.id)
            else:
                data = SPOTIFY.next(data)
            results += data['items']
        return [Playlist(playlist) for playlist in data]

    def is_followed(self) -> bool:
        return SPOTIFY.current_user_following_users([self.id])[0]

    def follow_user(self) -> None:
        return SPOTIFY.user_follow_users([self.id])

    def unfollow_user(self) -> None:
        return SPOTIFY.user_unfollow_users([self.id])



from typing import Dict, List, Literal, Optional

from melodine.utils import Image, URIBase
from melodine.configs import SPOTIFY
from melodine.models.base.artist import ArtistBase


class Artist(URIBase):
    '''
    Attributes
    ----------
        id: `str`
            spotify id for the album

        name: `str`
            name of the artist

        genres: `List[str]`
            a list of all the genres from an artist

        albums: `List[Album]`
            a list of all the albums from an artist

    methods
    -------
        artist_top_tracks: `List[Track]`
            a list of top tracks (a list containing track objects for the top tracks)
    '''

    __slots__ = (
        "followers",
        "genres",
        "images",
        "name",
        "href",
        "id",
        "uri",
        "_albums",
    )

    def __init__(self, data: Dict):
        self.id = data.get('id')  # pylint: disable=invalid-name
        self.uri = data.get('uri', None)
        self.href = data.get('external_urls').get('spotify', None)
        self.name = data.get('name', None)
        self.genres = data.get('genres', None)
        self.followers = data.get('followers', {}).get('total', None)
        self.images = [
            Image(**image)
            for image in data.get('images', [])
        ]

        self._albums = list()

    def __repr__(self) -> str:
        return f"melo.Artist - {(self.name or self.id or self.uri)!r}"

    @property
    def albums(self):
        if not self._albums:
            self.get_albums()
        return self._albums

    @property
    def total_albums(self):
        return SPOTIFY.artist_albums(self.id, limit=1)['total']

    def get_albums(
        self,
        limit: Optional[int] = 20,
        offset: Optional[int] = 0,
        album_type: Literal['album', 'single',
                            'appears_on', 'compilation'] = 'album'
    ) -> List:
        from .album import Album

        if len(self._albums) == 0:
            data = SPOTIFY.artist_albums(
                self.id, limit=limit, offset=offset, album_type=album_type)
            self._albums = list(Album(album) for album in data['items'])
        return self._albums

    def get_all_albums(self):
        from .album import Album

        offset = 0

        while len(self._albums) < self.total_albums:
            data = SPOTIFY.artist_albums(self.id, limit=50)

            offset += 50
            self._albums.extend(list(Album(album) for album in data['items']))
        return list(set(self._albums))

    def tracks(self):
        from .track import Track

        top = SPOTIFY.artist_top_tracks(self.id)
        return list(Track(track) for track in top['tracks'])

    def related_artists(self):
        related = SPOTIFY.artist_related_artists(self.id)
        return list(Artist(artist) for artist in related['artists'])



from typing import Dict, List

from melodine.configs import YTMUSIC
from melodine.models import ytmusic
from melodine.models.ytmusic.artist import Artist
from melodine.models.ytmusic.track import Track
from melodine.utils import Image, URIBase
from melodine.models.base.album import AlbumBase


class Album(URIBase):

    __slots__ = (
        '_data',
        '_id',
        '_name',
        '_href',
        'uri',
        '_audio_playlist_id',
        '_type',
        '_description',
        '_year',
        '_total_tracks',
        '_duration',
        '_images',
        '_artists',
        '_tracks',
    )

    def __init__(self, data: Dict) -> None:
        self._data = None

        self._id: str = data.get(
            'browseId',
            str()
        )
        self._name: str = data.get('title')
        self._href: str
        self.uri: str = f"ytmusic:album:{self.id}"

        self._audio_playlist_id: str = data.get('audioPlaylistId', None)

        self._type: str = data.get('type', None)
        self._description: str = data.get('description', None)
        self._year: str = data.get('year', None)
        self._total_tracks: int = data.get('trackCount', None)
        self._duration: int = data.get('duration_seconds', None)

        self._images: List[Image] = [
            Image(**image)
            for image in data.get('thumbnails', [])
        ]

        self._artists: List[Dict] = data.get('artists', [])
        self._tracks: List[Dict] = data.get('tracks', [])

    def _get_data(self) -> None:
        self._data = YTMUSIC.get_album(self.id)

    @property
    def id(self) -> str:
        if not self._id:
            self._id = YTMUSIC.get_album_browse_id(
                self._data.get('audioPlaylistId')
            )
        return self._id

    @classmethod
    def from_id(cls, id: str) -> "Album":
        # the given id could be a playlistsId or a browseId.
        # a browseId starts with 'MPREb' and a audioPlaylistId starts with 'OLAK5uy'
        return cls(data={'browseId' if id.startswith("MPREb") else 'audioPlaylistId': id})

    @classmethod
    def partial(cls, data: Dict) -> "Album":
        """
        for data obtained from tracks and other shorter sources.
            `{'name': '<album-name>', 'id': '<album-id>'}`
        """
        return cls(data={
            'title': data['name'],
            'browseId': data['id']
        })

    @property
    def name(self) -> str:
        if self._name is None:
            if self._data is None:
                self._get_data()
            self._name = self._data['title']
        return self._name

    @property
    def href(self) -> str:
        return "https://music.youtube.com/playlist?list=" + self.audio_playlist_id

    @property
    def audio_playlist_id(self) -> str:
        if self._audio_playlist_id is None:
            if self._data is None:
                self._get_data()
            self._audio_playlist_id = self._data.get('audioPlaylistId')
        return self._audio_playlist_id

    @property
    def type(self) -> str:
        if self._type is None:
            if self._data is None:
                self._get_data()
            self._type = self._data.get('type')
        return self._type

    @property
    def description(self) -> str:
        if self._description is None:
            if self._data is None:
                self._get_data()
            self._description = self._data.get('description')
        return self._description

    @property
    def year(self) -> str:
        if self._year is None:
            if self._data is None:
                self._get_data()
            self._year = self._data.get('year')
        return self._year

    @property
    def total_tracks(self) -> int:
        if self._total_tracks is None:
            if self._data is None:
                self._get_data()
            self._total_tracks = self._data.get('trackCount')
        return self._total_tracks

    @property
    def duration(self) -> int:
        if self._duration is None:
            if self._data is None:
                self._get_data()
            self._duration = self._data.get('duration_seconds')
        return self._duration

    @property
    def images(self) -> List[Image]:
        if len(self._images) == 0:
            if self._data is None:
                self._get_data()
            self._images = [
                Image(**image)
                for image in self._data['thumbnails']
            ]
        return self._images

    @property
    def artists(self) -> List["ytmusic.Artist"]:
        if len(self._artists) == 0:
            if self._data is None:
                self._get_data()
            self._artists = self._data.get('artists')
        for idx, artist in enumerate(self._artists.copy()):
            if not isinstance(artist, Artist):
                artist = Artist.partial(artist)
                self._artists.insert(idx, artist)
                del self._artists[idx + 1]
        return self._artists

    @property
    def tracks(self) -> List["ytmusic.Track"]:
        if len(self._tracks) == 0:
            if self._data is None:
                self._get_data()
            self._tracks = self._data.get('tracks')
        for idx, track in enumerate(self._tracks.copy()):
            if not isinstance(track, Track):
                track = Track(track, album=self)
                self._tracks.insert(idx, track)
                del self._tracks[idx + 1]
        return self._tracks



from typing import Dict

from melodine.utils import URIBase


class Device(URIBase):
    def __init__(self, data: Dict) -> None:
        self.id: str = data.get('id')
        self.name: str = data.get('name')
        self.type: str = data.get('type')

        self.volume: int = data.get('volume')

        self.is_active: bool = data.get('is_active')
        self.is_private_session: bool = data.get('is_private_session')
        self.is_restricted = data.get('is_restricted')

    def __repr__(self):
        return f"<spotify.Device: {(self.name or self.id)!r}>"



from abc import ABCMeta


class ArtistBase(metaclass=ABCMeta):
    @classmethod
    def __subclasshook__(cls, subclass):
        return (
            hasattr(subclass, 'id') and
            hasattr(subclass, 'name') and
            hasattr(subclass, 'href') and
            hasattr(subclass, 'uri') and
            hasattr(subclass, 'images') and

            hasattr(subclass, 'album') and
            hasattr(subclass, 'url') and

            hasattr(subclass, 'related_artists') and callable(
                subclass.related_artists)
        )

    # @property
    # @abstractmethod
    # def id(self):
    #     raise NotImplementedError

    # @property
    # @abstractmethod
    # def name(self):
    #     raise NotImplementedError

    # @property
    # @abstractmethod
    # def href(self):
    #     raise NotImplementedError

    # @property
    # @abstractmethod
    # def uri(self):
    #     raise NotImplementedError

    # @property
    # @abstractmethod
    # def images(self):
    #     raise NotImplementedError

    # @property
    # @abstractmethod
    # def followers(self):
    #     raise NotImplementedError

    # @property
    # @abstractmethod
    # def albums(self):
    #     raise NotImplementedError

    # @property
    # @abstractmethod
    # def tracks(self):
    #     raise NotImplementedError

    # @property
    # @abstractmethod
    # def related_artists(self):
    #     raise NotImplementedError



import json
import os
# import pickle

import appdirs
import spotipy
from ytmusicapi import YTMusic
from youtube_dl import YoutubeDL
import pyyoutube


APP_DIR = os.path.join(appdirs.user_data_dir(), '.melo')
CACHE_PATH = os.path.join(APP_DIR, 'spotify_cache')
PICKLES_DIR = os.path.join(APP_DIR, 'pickles')
TEMPFILES_DIR = os.path.join(APP_DIR, 'tempfiles')

# if not os.path.exists(PICKLES_DIR):
#     os.mkdir(PICKLES_DIR)

# if not os.path.exists(os.path.join(PICKLES_DIR, 'YT.pickle')):
#     import pyyoutube
#     YT = pyyoutube.Api(api_key='AIzaSyCq47Zxsu4pN1MMWBNa04380TGDxT7hrQM')
#     with open(os.path.join(PICKLES_DIR, 'YT.pickle'), 'wb') as file:
#         pickle.dump(YT, file, protocol=-1)
# else:
#     with open(os.path.join(PICKLES_DIR, 'YT.pickle'), 'rb') as file:
#         YT: "pyyoutube.Api" = pickle.load(file)


YT = pyyoutube.Api(api_key='AIzaSyCq47Zxsu4pN1MMWBNa04380TGDxT7hrQM')
YTDL_CONFIG = {'quiet': True}
# if not os.path.exists(os.path.join(PICKLES_DIR, 'YTDL.pickle')):
#     from youtube_dl import YoutubeDL
#     YTDL = YoutubeDL(YTDL_CONFIG)
#     with open(os.path.join(PICKLES_DIR, 'YTDL.pickle'), 'wb') as file:
#         pickle.dump(YoutubeDL, file, protocol=-1)
# else:
#     with open(os.path.join(PICKLES_DIR, 'YTDL.pickle'), 'rb') as file:
#         YTDL = pickle.load(file)(YTDL_CONFIG)

YTDL = YoutubeDL(YTDL_CONFIG)

SCOPES = '''
            user-read-playback-state
            user-follow-read
            user-follow-modify
            playlist-read-private
            playlist-read-collaborative
            playlist-modify-private
            playlist-modify-public
            user-read-recently-played
            user-library-read
            user-library-modify
            user-top-read
            user-read-private
        '''  # pylint: disable=invalid-name

if os.path.exists(CACHE_PATH):
    SPOTIFY = spotipy.Spotify(
        auth_manager=spotipy.SpotifyOAuth(
            scope=SCOPES,
            client_id="22e27810dff0451bb93a71beb5e4b70d",
            client_secret="6254b7703d8540a48b4795d82eae9300",
            redirect_uri="http://localhost:8080/",
            cache_handler=spotipy.CacheFileHandler(
                cache_path=CACHE_PATH
            )
        )
    )
else:
    SPOTIFY = spotipy.Spotify(
        auth_manager=spotipy.SpotifyClientCredentials(
            client_id='22e27810dff0451bb93a71beb5e4b70d',
            client_secret='6254b7703d8540a48b4795d82eae9300'
        )
    )

_ytmusic_cookies = {
    "accept": "*/*",
    "accept-encoding": "gzip, deflate",
    "accept-language": "en-US,en;q=0.9",
    "authorization": "SAPISIDHASH 1643090086_74cc0985095757af3f1af65902686e81b6637950",
    "content-encoding": "gzip",
    "content-type": "application/json",
    "cookie": "VISITOR_INFO1_LIVE=uyP5rRxNQhk; PREF=tz=Asia.Calcutta&f6=40000000; YSC=Q9aJ39XjaxI; SID=GQiVhB2tjMF5Z0BE6-SY--IlCwcC_z3cS4W8BGpxVd7stlEzvK9hxJn9F7fJub0alxpgdA.; __Secure-1PSID=GQiVhB2tjMF5Z0BE6-SY--IlCwcC_z3cS4W8BGpxVd7stlEzHz9XaSut4DXx5clU4qpx1A.; __Secure-3PSID=GQiVhB2tjMF5Z0BE6-SY--IlCwcC_z3cS4W8BGpxVd7stlEzYr0_Hswrcd9rmYm2Xxia_Q.; HSID=AXeZ83FjzhA8x1gY2; SSID=A0Z7pkk9iPB6BX-3f; APISID=04432aWLTO54P4Gp/A-Oni_5vP1OAfIYeT; SAPISID=Orxl6q7KyzDWyxMB/Ag-yFYBZUXWy4E778; __Secure-1PAPISID=Orxl6q7KyzDWyxMB/Ag-yFYBZUXWy4E778; __Secure-3PAPISID=Orxl6q7KyzDWyxMB/Ag-yFYBZUXWy4E778; LOGIN_INFO=AFmmF2swRAIgHSN980c9f8lkxpke8TdftmbblS4jZ7yTp1YTiR4DLvsCICKFB1SIMxehNygfMgnGc1YN6vED-kFyi8MzTWG0SyfM:QUQ3MjNmejk0emRXVURCMmsyLXNxaGN2LUN3N2FfOVkyNHlDR3lDb0poVmhMV0xDMnVRQUdiaHlueWVkSTVNZHFNejY5SWdiVFJ6NTBNbGNuU2o3UXNjazRRMmVDYmlTZ3Zrc1pMZWF0ZE5oV0xzeWpFQXJKWHlHcXRESFIxQnAzX3VMT1FJOElUOHNFSHN5ZXZNaUNBMzk2TXB2U1YtLXNR; SIDCC=AJi4QfGNmCEAfa_dQq_4mIiapY6nuREkM-CXP4G7dphqBk6k8wJgqW538_oPf3A8yMrd7zkGCA; __Secure-3PSIDCC=AJi4QfEs1oqCsKgL0dbAqFDExHQDcWnX49JGwoqdKv0eo_1NxlzcNAYsNQ_Rv4Vd3l32kda0Vw",
    "dnt": "1",
    "origin": "https://music.youtube.com",
    "referer": "https://music.youtube.com/",
    "sec-ch-ua": "\" Not;A Brand\";v=\"99\", \"Google Chrome\";v=\"97\", \"Chromium\";v=\"97\"",
    "sec-ch-ua-arch": "\"x86\"",
    "sec-ch-ua-bitness": "\"64\"",
    "sec-ch-ua-full-version": "\"97.0.4692.71\"",
    "sec-ch-ua-mobile": "?0",
    "sec-ch-ua-platform": "\"Windows\"",
    "sec-ch-ua-platform-version": "\"14.0.0\"",
    "sec-fetch-dest": "empty",
    "sec-fetch-mode": "same-origin",
    "sec-fetch-site": "same-origin",
    "user-agent": "Mozilla/5.0 (Windows NT 10.0; Win64; x64; rv:88.0) Gecko/20100101 Firefox/88.0",
    "x-client-data": "CIq2yQEIo7bJAQjBtskBCKmdygEI7enKAQjq8ssBCJ75ywEI1/zLAQjmhMwBCIOVzAEI9pXMAQ==",
    "x-goog-authuser": "0",
    "x-goog-visitor-id": "Cgt1eVA1clJ4TlFoayikob6PBg%3D%3D",
    "x-origin": "https://music.youtube.com",
    "x-youtube-client-name": "67",
    "x-youtube-client-version": "1.20220119.00.00"
}

YTMUSIC = YTMusic(auth=json.dumps(_ytmusic_cookies))

if os.path.exists(TEMPFILES_DIR):
    os.rmdir(TEMPFILES_DIR)



# region Imports
import curses

import math
import os
import threading
import time
from curses import panel, wrapper
from curses.textpad import Textbox, rectangle

from ffpyplayer.player import MediaPlayer
from ffpyplayer.tools import loglevels, set_log_callback
from melodine import innertube, spotify, utils, ytmusic
from melodine.configs import APP_DIR
from melodine.player import Player
# endregion

# region to make ffpyplayer not output anything
loglevel_emit = 'error'

def log_callback(message, level):
    if message and loglevels[level] <= loglevels[loglevel_emit]:
        print ("error")


set_log_callback(log_callback)
# endregion

# region God Help Me


global ffs
class Menu():
    search_type = "tracks"
    with open(os.path.join(APP_DIR,"ffs.txt"), 'w') as ffs:
        ffs.write(search_type)

    def __init__(self, items, stdscreen):
        self.window = stdscreen.subwin(5, 12, 0, width - 13)
        self.window.keypad(1)
        self.panel = panel.new_panel(self.window)
        self.panel.hide()
        panel.update_panels()

        self.position = 0
        self.items = items

    def navigate(self, n):
        self.position += n
        if self.position < 0:
            self.position = 0
        elif self.position >= len(self.items):
            self.position = len(self.items) - 1

    def display(self):
        self.panel.top()
        self.panel.show()
        self.window.clear()

        while True:
            self.window.refresh()
            curses.doupdate()
            for index, item in enumerate(self.items):
                if index == self.position:
                    mode = curses.A_REVERSE
                else:
                    mode = curses.A_NORMAL

                msg = "%d. %s" % (index, item[0])
                self.window.addstr(1 + index, 1, msg, mode)

            key = self.window.getch()

            if key in [curses.KEY_ENTER, ord("\n")]:
                search_type = self.items[self.position][0].lower()

                with open(os.path.join(APP_DIR,"ffs.txt"), 'w') as ffs:
                    ffs.write(search_type)
                break

            elif key == curses.KEY_UP:
                self.navigate(-1)

            elif key == curses.KEY_DOWN:
                self.navigate(1)

        self.window.clear()
        self.panel.hide()
        panel.update_panels()
        curses.doupdate()
# endregion

def runner(stdscr):

    curses.noecho()
    stdscr.keypad(True)
    global height
    global width
    global playing
    global player
    global index
    index = 0
    playing = False
    height, width = stdscr.getmaxyx()

    # print(height)

    # Quits if terminal is too small
    if height < 14 or width < 40:
        curses.nocbreak()
        stdscr.keypad(False)
        curses.echo()
        curses.endwin()

        # print("Terminal too small!")

        quit()
    curses.init_pair(1, curses.COLOR_GREEN, curses.COLOR_BLACK)
    curses.init_pair(2, curses.COLOR_MAGENTA, curses.COLOR_BLACK)
    # Initialises search box and player window
    box = draw_search_box(stdscr)
    playerwin = draw_player(stdscr)
    menu = draw_type_menu(stdscr)
    typewin = stdscr.subwin(3, 9, 1, width - 13)
    typewin.border()
    typewin.addstr(1, 1, menu.search_type)
    typewin.refresh()
    playerwin.refresh()
    curses.curs_set(0)
    while True:
        c = stdscr.getch()
        # If window resized redraw elements
        # FIXME: Does not redraw correctly
        if c == curses.KEY_RESIZE:
            if height < 14 or width < 40:
                curses.nocbreak()
                stdscr.keypad(False)
                curses.echo()
                curses.endwin()

                # print("Terminal too small!")

                quit()
            if box.gather():
                results = search(box, stdscr, menu)
            box = draw_search_box(stdscr)
            playerwin = draw_player(stdscr)
            playerwin.refresh()

        elif c == ord('/'):
            results = search(box, stdscr, menu)
        elif c == ord('q'):
            curses.nocbreak()
            stdscr.keypad(False)
            curses.echo()
            curses.endwin()
            break
        elif c >= ord('0') and c <= ord('9'):

            selected = results[int(chr(c))]
            # utils.put_notification(selected)
            player_update(playerwin, selected.artists[0].name, selected.name, selected.duration)
            # url = innertube.InnerTube().player(selected.id)['streamingData']['formats'][-1]['url']
            player = Player()
            player.play(selected)
            global bar
            bar = threading.Thread(target=progressbar, args=(playerwin, selected.duration, player)).start()

        elif c == ord('p'):
            if player:
                player.toggle_pause()
                player_update(playerwin)
                bar = threading.Thread(target=progressbar, args=(playerwin, selected.duration, player)).start()
        elif c == ord('t'):
            menu.display()

            typewin = stdscr.subwin(3, 9, 1, width - 13)
            typewin.border()
            ffs = ""
            with open(os.path.join(APP_DIR,"ffs.txt"), "r") as help:
                ffs = help.read()
            typewin.addstr(1, 1, ffs)
            typewin.refresh()
        elif c == ord("l"):
            with open(os.path.join(APP_DIR, "Liked.txt")) as liked:
                if selected.name not in liked.readlines():
                    liked.write(selected.name)

        elif c == curses.KEY_UP:
            with open(os.path.join(APP_DIR,"History.txt"), "r") as history:
                search_term = history.readlines()[index]
                results = search(box, stdscr, menu, search_term)
                stdscr.addstr(2, 1, search_term[:30])
            index += 1

# region Curses Functions
# Initialises player window with "Nothing Playing"

def draw_player(stdscr):
    playerwin = curses.newwin(5, width - 1, height - 5, 0)
    playerwin.border()
    playerwin.addstr(0, 1, "Nothing Playing", curses.color_pair(1) | curses.A_BOLD)

    playerwin.refresh()
    playing = False
    return playerwin

# Method to draw panel for search type

def draw_type_menu(stdscr):
    menu_items = [("Tracks", curses.flash), ("Artists", curses.flash), ("Albums", curses.flash)]

    menu = Menu(menu_items, stdscr)
    return menu

# Method to draw search box

def draw_search_box(stdscr):
    stdscr.addstr(0, 0, "Search (/ to focus):", curses.color_pair(1) | curses.A_BOLD)
    stdscr.keypad(True)
    editwin = curses.newwin(1,30, 2,1)
    rectangle(stdscr, 1,0, 1+1+1, 1+30+1)

    stdscr.refresh()
    box = Textbox(editwin)
    return box

# Updates player window with track data and progressbar

def player_update(player, artist=None, title=None, duration=None):
    minutes = math.floor(duration / 60)
    seconds = math.floor(duration % 60)
    playerwin = curses.newwin(5, width - 1, height - 5, 0)
    length = width - 16
    playerwin.border()
    if playing == True:
        playerwin.addstr(0, 1, "Paused", curses.color_pair(1) | curses.A_BOLD)
        playing == False
    else:
        playerwin.addstr(0, 1, "Playing", curses.color_pair(1) | curses.A_BOLD)
        playing == True
    if artist and title:
        playerwin.addstr(1, 2, title, curses.color_pair(2) | curses.A_BOLD)
        playerwin.addstr(2, 2, artist, curses.color_pair(1))
        playerwin.addstr(3, 2, "0:00")
        playerwin.addstr(3, width - 7, f"{minutes}:{'{:02d}'.format(seconds)}")
        playerwin.addstr(3, 8, "-" * length)
    playerwin.refresh()

# Updates progressbar

def progressbar(playerwin, duration, player):
    '''pts = 0
    while True:
        if pts == duration:
            break
        length = width - 16
        percent = duration / length
        position = math.floor(pts / percent)
        # playerwin.addstr(1, 30, f"dur: {duration}%: {round(percent,2)}, len: {round(length,2)}, pts%%: {round(pts % percent, 2)}, pos: {position}, pts: {'{:03d}'.format(pts)}")
        playerwin.addstr(3, 8, "=" * position)
        minutes = math.floor(pts / 60)
        seconds = math.floor(pts % 60)
        playerwin.addstr(3, 2, f"{minutes}:{'{:02d}'.format(seconds)}")
        playerwin.refresh()
        pts += 1
        time.sleep(1)'''
    pts = player.get_current_timestamp()
    while playing and pts != duration:
        length = width - 16
        percent = duration / length
        position = math.floor(pts / percent)
        # playerwin.addstr(1, 30, f"dur: {duration}%: {round(percent,2)}, len: {round(length,2)}, pts%%: {round(pts % percent, 2)}, pos: {position}, pts: {'{:03d}'.format(pts)}")
        playerwin.addstr(3, 8, "=" * position)
        minutes = math.floor(pts / 60)
        seconds = math.floor(pts % 60)
        playerwin.addstr(3, 2, f"{minutes}:{'{:02d}'.format(seconds)}")
        playerwin.refresh()
        time.sleep(1)

def search(box, stdscr, menu, search=None,):
    curses.curs_set(1)
    if not search:
        box.edit()
        # Get resulting contents
        search = box.gather()
        # Print the window to the screen

    with open(os.path.join(APP_DIR,"ffs.txt"), "r") as help:
            ffs = help.read()
    search_type = ffs
    results = spotify.search(search, types=[search_type])

    color = curses.has_colors()
    # Print results based on screensize, maxed at 10
    result_num = height - 9
    if height - 9 > 10:
        result_num = 10
    result_box = curses.newwin(result_num + 2, width - 1, 4, 0)
    result_box.border()

    result_box.addstr(0, 1, "Enter index number to play", curses.color_pair(1) | curses.A_BOLD)
    line = 1

    if search_type == "tracks":
        results = results.tracks
    elif search_type == "artists":
        results = results.artists
    elif search_type == "albums":
        results = results.albums
    for result in results[:result_num]:
        if search_type == "tracks":
            name = str(result.name)
            artists_list = []
            for artist in result.artists:
                artists_list.append(str(artist.name))
            artists = ' & '.join(artists_list)
            if len(artists) > width/3:
                artists = result.artists[0].name
            if len(name) > width/3:
                name = name[:(int(width/3))]

            duration_minute, duration_seconds = math.floor(result.duration / 60), "{:02d}".format(math.floor(result.duration % 60))
            result_box.addstr(line, 1, f"{line - 1}. |{name}{(int(width/3) - len(name)) * ' '}|{artists}{(int(width/3) - len(artists)) * ' '}|{duration_minute}:{duration_seconds}")

            line += 1
            curses.curs_set(0)
            result_box.refresh()
        elif search_type == "artists":
            name = str(result.name)
            if len(name) > width/3:
                name = name[:(int(width/3))]
            result_box.addstr(line, 1, f"{line - 1}. |{name}")
            line += 1
            curses.curs_set(0)
            result_box.refresh()
        elif search_type == "albums":
            name = str(result.name)
            artists_list = []
            for artist in result.artists:
                artists_list.append(str(artist.name))
            artists = ' & '.join(artists_list)
            if len(artists) > width/3:
                artists = result.artists[0].name
            if len(name) > width/3:
                name = name[:(int(width/3))]

            duration_minute, duration_seconds = math.floor(result.duration / 60), "{:02d}".format(math.floor(result.duration % 60))
            result_box.addstr(line, 1, f"{line - 1}. |{name}{(int(width/3) - len(name)) * ' '}|{artists}{(int(width/3) - len(artists)) * ' '}|{duration_minute}:{duration_seconds}")

            line += 1
            curses.curs_set(0)
            result_box.refresh()
    return results

# endregion

if __name__ == "__main__":
    wrapper(runner)


from melodine import player
from melodine.models import (
    base,
    spotify,
    ytmusic,
    youtube
)

__all__ = (
    "base",
    "spotify",
    "ytmusic",
    "youtube",
    'player',
)

from melodine.configs import APP_DIR
import os
if not os.path.exists(APP_DIR):
    os.mkdir(APP_DIR)



from abc import ABCMeta


class TrackBase(metaclass=ABCMeta):
    @classmethod
    def __subclasshook__(cls, subclass):
        return (
            hasattr(subclass, 'id') and
            hasattr(subclass, 'name') and
            hasattr(subclass, 'href') and
            hasattr(subclass, 'uri') and
            hasattr(subclass, 'explicit') and
            hasattr(subclass, 'duration') and
            hasattr(subclass, 'images') and

            hasattr(subclass, 'artists') and
            hasattr(subclass, 'album') and
            hasattr(subclass, 'url') and

            hasattr(subclass, 'get_recommendations') and callable(
                subclass.get_recommendations)
        )

    # @property
    # @abstractmethod
    # def id(self):
    #     raise NotImplementedError

    # @property
    # @abstractmethod
    # def name(self):
    #     raise NotImplementedError

    # @property
    # @abstractmethod
    # def href(self):
    #     raise NotImplementedError

    # @property
    # @abstractmethod
    # def uri(self):
    #     raise NotImplementedError

    # @property
    # @abstractmethod
    # def explicit(self):
    #     raise NotImplementedError

    # @property
    # @abstractmethod
    # def duration(self):
    #     raise NotImplementedError

    # @property
    # @abstractmethod
    # def images(self):
    #     raise NotImplementedError

    # # @classmethod
    # # @abstractmethod
    # # def from_id(cls, id: str) -> "TrackBase":
    # #     raise NotImplementedError

    # @property
    # @abstractmethod
    # def artists(self):
    #     raise NotImplementedError

    # @property
    # @abstractmethod
    # def album(self):
    #     raise NotImplementedError

    # @property
    # @abstractmethod
    # def url(self):
    #     raise NotImplementedError

    # @abstractmethod
    # def get_recommendations(self):
    #     raise NotImplementedError



# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['melodine',
 'melodine.models',
 'melodine.models.base',
 'melodine.models.spotify',
 'melodine.models.youtube',
 'melodine.models.ytmusic',
 'melodine.player']

package_data = \
{'': ['*']}

install_requires = \
['appdirs>=1.4.4,<2.0.0',
 'ffpyplayer>=4.3.5,<5.0.0',
 'python-youtube>=0.8.3,<0.9.0',
 'requests>=2.28.1,<3.0.0',
 'spotipy>=2.21.0,<3.0.0',
 'typing-extensions>=3.7.4,<4.0.0',
 'youtube_dl>=2021.12.17,<2022.0.0',
 'ytmusicapi>=0.24.0,<0.25.0']

setup_kwargs = {
    'name': 'melodine',
    'version': '0.2.3',
    'description': 'cohesive models for seamlessly interacting with the Spotify and YouTube Music APIs',
    'long_description': '# melodine\n\nmelo provides consistent data models over Spotify and YouTube Music API objects.\nit\'s main purpose is to model objects from both platforms in a way that requires no distinction based on it\'s source while using them\n\nfor example, a Track object from melo provides the same properties and behaviour irrespective of wether the concerned track is from Spotify or YouTube Music.\n\n# Installation\n\n    pip install melodine\n\n# Usage and Features\n\n## Consistent Data Models\n\nthe term "consistent" here implies that the models work the same way without having to care about it\'s source.\n\nThe Spotify and YouTube Music API provide similar attributes related to objects like a song\'s title, the artists related to the song, the album it\'s from and so on. However their schemas and the way in which objects interrelate wiht each other is totally different. Thus it becomes a pain when integrating both services into a singe application.\n\nmelo tries to solve this problem by internally separating the modelling of objects from both the services with the same structure and exporting them as global melo models, thus resulting in a robust abstraction over both services.\n\nthat means that a YouTube Music track and Spotify track can be handled as the same global melo Track object. This is done by having common attributes between the two separate internal models, like a track\'s name, it\'s artists and the album it\'s from, etc. such gloabl attributes that all melo tracks have (wether from Spotify or YouTube Music) are -\n  \n| attribute           | description                                                                               |\n|---------------------|-------------------------------------------------------------------------------------------|\n| album               | the melo album object for a track                                                         |\n| artists             | the melo artist object for a track                                                        |\n| name                | the name of the track                                                                     |\n| duration            | duration of the track (in seconds)                                                        |\n| href                | link to the track\'s source page                                                           |\n| id                  | the id of the track (the video for the track on ytmusic)                                  |\n| uri                 | a uri of the format `source:type:id` for example - `spotify:track:0hz0bTQC2VVb4CEjLxmKiH` |\n| url                 | the url to the song playback for streaming                                                |\n| images              | a list of `Image` objects (the cover art for the track)                                   |\n| get_recommendations | a getter for track recommendations related to this track                                  |\n\n_Note: there are some extra attributes for a spotify track, but are irrelevant to the core attributes_\n\nthe same applies to other models as well (artist, album, playlists, vidoes, and shows, etc)\n\n## search\n\nmelo allows searching content from both services. Results from specific source or of specific types can be fetched.\n\n### Spotify\n\n```py\nimport melodine as melo\n\nresults = melo.spotify.search(\'Martin Garrix\')\n```\n\n### YT Music\n\n```py\nimport melodine as melo\n\nresults = melo.ytmusic.search(\'Martin Garrix\')\n```\n\nthe `search` function fetches results based on the model used. It returns a `SearchResults` instance. separated the search results based on result types. for example, only tracks results can be accessed as `SearchResults.tracks` which returns an array of Track objects (an empty array if there\'s no tracks in the results).\n\n```py\nresults = melo.spotify.search(\'sewerslvt\') \n\nprint(results.tracks)\n\n# [<melo.Track - \'Cyberia Lyr1\'>, <melo.Track - \'Ecifircas\'>, <melo.Track - \'goodbye\'>]\n```\n\nto fetch specific types of results\n\n```py\nimport melo\n\nresults = melo.spotify.search(\'sewerslvt\', types=[\'track\', \'playlist\'])  \n\nprint(results.playlists)\n\n# [<melo.Playlist - \'This Is Sewerslvt\'>, <melo.Playlist - \'Breakcore Heaven\'>]\n\nprint(results.tracks)   \n\n# [<melo.Track - \'Ecifircas\'>, <melo.Track - \'goodbye\'>, <melo.Track - \'Newlove\'>]\n\nprint(results.albums)\n\n# []\n```\n\nOnly the specified types of results are fetched and the other fields remain empty.\n\n```py\nimport melo\n\nytsearch = melo.ytmusic.search(\'sewerslvt\', source=[\'ytmusic\'], types=[\'album\']) \n\nytsearch.albums  \n\n# [melo.Album - \'Sewer//slvt\', melo.Album - "we had good times together, don\'t forget that"]\n```\n\nOnly YouTube Music albums will be fetched. Any combination of parameters can be used as per convinience.\n\n## Nested Models\n\nmelo models are connected to each other. a track object has an Artist object as it\'s artist parameter, that artist in turn has it\'s own top tracks, albums and those albums have the tracks in them which are fully fledged track objects themselves which means they can lead to other recommended tracks.\n\nas extensive as it gets, an artist could also lead to other similar artists with their own tracks tracks and albums. there\'s a lot of exploring to do out there\n\nit\'s understandable if all thats too mind boggling, here\'s a code example\n\n```py\nimport melo\n\nresults = melo.spotify.search(\'potsu\', source=[\'ytmusic\'], types=[\'artist\'])\n\ntrack_name = results.artists[0].albums[3].get_tracks()[0].name  # \'bird\'\n```\n\nthis crazy chaining implies getting the name of the 1st track from an artist\'s 3rd album where the artist is the 1st search result for a search term.\n\neach step in fetching the desired metric is done lazily which implies melo\'s idea usage for using in TUI application where details need to be loaded only on clicks or other interactions.\n\n## Spotify Authorization\n\nmelo can be used to access data to a user\'s spotify data. the user needs to provide consent to the developer to be able to use their spotify data.\n\nThe purpose for including this functionality is to be able to use melo\'s models with a user\'s spotify library which includes lot of liked tracks, saved albums, artists and curated playlists to extend it to the user\'s content.\n\nto invoke spotify authorization\n\n```py\nimport melo\n\nmelo.spotify.client.authorize()\n```\n\nthis is a one-time action and does not need to be repeated once fulfilled. it is this way in order to emulate usage in a CLI or TUI application where user signin needs to be done just once.\n\nOnce authorized, the `melo.spotify.client` object can be used to make authenticated requests to the Spotify API to get the user\'s liked playlsits, saved albums & artists, a user\'s top & recently played track, and much more.\n\n```py\nfrom melo import spotify.client as client\n\n# assumes the client is already authorized\nrecent_tracks = client.recently_played()\n\n# [<melo.Track - \'Star Shopping\'>, <melo.Track - \'Rum & Her\'>, <melo.Track - \'違う\'>]\n```\n\n# Planned Features\n\n- Implementing YouTube OAuth to retrieve a user\'s YouTube playlists\n- fetching the lyrics / captions for a track\n- transferring spotify playlists to youtube / youtube music and vice versa\n',
    'author': 'addyett',
    'author_email': 'g.aditya2048@gmail.com',
    'maintainer': 'None',
    'maintainer_email': 'None',
    'url': 'None',
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'python_requires': '>=3.6,<4.0',
}


setup(**setup_kwargs)



from typing import Dict, List, Union

from melodine.utils import URIBase, Image
from melodine.configs import YTMUSIC
from melodine.innertube import InnerTube
from melodine.models import ytmusic
from melodine.models.ytmusic.artist import Artist
from melodine.models.ytmusic.track import Track


class Video(URIBase):

    __slots__ = (
        '_data',
        'id',
        '_name',
        'href',
        'uri',
        '_artists',
        '_explicit',
        '_views',
        '_duration',
        '_images',

        '_url',
        '_recs',
        '_recs_offset'
    )

    def __init__(self, data: Dict) -> None:
        self._data = None

        self.id: str = data.get('videoId', data.get('resourceId'))
        self._name: str = data.get('title', None)
        self.href: str = "https://music.youtube.com/watch?v=" + self.id
        self.uri: str = "ytmusic:video:" + self.id

        self._artists: List[Dict] = data.get('artists', [])

        self.explicit: bool = data.get('isExplicit', False)
        self._views: str = data.get('views', data.get('viewCount', None))

        self._duration: Union[int, None] = data.get(
            'lengthSeconds', data.get('duration_seconds', data.get('length', None)))

        self._images: List[Image] = [
            Image(**image)
            for image in data.get('thumbnails', [])
        ]

        self.playlist_id = data.get('playlist_id', None)

        self._url = str()
        self._recs = list()
        self._recs_offset = int()  # aka 0

    def _get_data(self) -> None:
        self._data = YTMUSIC.get_song(self.id)['videoDetails']

    @classmethod
    def from_id(cls, id) -> "Video":
        return cls(data={'videoId': id})

    @classmethod
    def from_raw(cls, data: Dict) -> "Video":
        data['artists'] = [
            {
                'name': data.pop('author'),
                'id': data.pop('channelId')
            }
        ]
        data['thumbnails'] = data.pop('thumbnail').get('thumbnails')
        return cls(data=data)

    @property
    def name(self) -> str:
        if self._name is None:
            if self._data is None:
                self._get_data()
            self._name = self._data['title']
        return self._name

    @property
    def artists(self) -> Union[Artist, "ytmusic.User"]:
        from .user import User

        if len(self._artists) == 0:
            if self._data is None:
                self._get_data()
            self._artists = [{'name': self._data.pop(
                'author'), 'id': self._data.pop('channelId')}]
        for idx, artist in enumerate(self._artists.copy()):
            if not isinstance(artist, (Artist, User, )):
                # since a video is'nt neccessarily owned by an artist each time,
                # so it's from a user in case it's not from an artist.
                try:
                    user = YTMUSIC.get_user(artist['id'])
                    artist = User(user, id=artist['id'])
                except (KeyError, ValueError):
                    artist = Artist.partial(artist)
                self._artists.insert(idx, artist)
                del self._artists[idx + 1]
        return self._artists

    @property
    def views(self) -> int:
        if self._views is None or not self._views.isdigit():
            if self._data is None:
                self._get_data()
            self._views = self._data['viewCount']
        return int(self._views)

    @property
    def duration(self) -> int:
        if self._duration is None:
            if self._data is None:
                self._get_data()
            self._duration = self._data['lengthSeconds']
        if isinstance(self._duration, str):
            if ':' in self._duration:
                self._duration = self._duration.split(':')
                self._duration = int(
                    self._duration[0]) * 60 + int(self._duration[1])
            else:
                self._duration = int(self._duration)
        return self._duration

    @property
    def images(self) -> List[Image]:
        if len(self._images) == 0:
            if self._data is None:
                self._get_data()
            self._images = self._data.get('thumbnail')['thumbnails']
        for idx, image in enumerate(self._images.copy()):
            if not isinstance(image, Image):
                image = Image(**image)
                self._images.insert(idx, image)
                del self._images[idx + 1]
        return self._images

    @property
    def url(self) -> str:
        '''Get the playback URL for a track or video

        retrieves the higest quality adaptive track by default
        '''
        if self._url:
            return self._url

        innertube = InnerTube()
        video_info = innertube.player(self.id)
        self._url = video_info['streamingData']['adaptiveFormats'][-1]['url']

        return self._url

    def cache_url(self) -> None:
        '''just a dummy call to trigger the url fetching.'''
        if not self.url:
            self._url = self.url

    @property
    def recommendations(self) -> List["Track"]:
        '''A list of recommended tracks for this track already fetched upto this point.

        it is initially empty and gets populated as more calls are made to `get_recommendations` are made.

        use `get_recommendations` to fetch new recommendations. instead'''
        return self._recs

    def get_recommendations(self, limit: int = 10) -> List["Track"]:
        '''used to get new recommendations for the track.

        earlier fetched recommended tracks can be accessed whith `.recommendations'''

        data = YTMUSIC.get_watch_playlist(
            self.id,
            f"RDAMVM{self.id}",
            limit=self._recs_offset + limit
        )

        tracks = data['tracks'][self._recs_offset: (self._recs_offset + limit)]
        tracks = [Track(track) for track in tracks]
        self._recs.append(tracks)
        self._recs_offset += limit

        return tracks



from typing import Dict, List, Optional, Union

from melodine.models import spotify

from melodine.utils import Image, URIBase
from melodine.configs import SPOTIFY
from melodine.models.spotify.track import PlaylistTrack, Track
from melodine.models.base.playlist import PlaylistBase


class Playlist(URIBase):  # pylint: disable=too-many-instance-attributes
    '''A Spotify playlist object'''
    __slots__ = (
        "tracks",
        "total_tracks",
        "images",
        "href",
        "name",
        "owner",
        "snapshot_id",
        "public",
        "collaborative",
        "description",
        "fllowers",
        "uri",
        "id",
    )

    def __init__(self, data: Dict) -> None:
        from .user import User

        self.id: str = data.get('id')  # pylint: disable=invalid-name

        self.name: str = data.get('name')
        self.owner: User = User(data.get('owner'))
        self.snapshot_id: str = data.get('snapshot_id')
        self.href: str = 'https://open.spotify.com/playlist/' + self.id
        self.uri: str = data.get('uri')

        self.public: bool = data.get('public')
        self.collaborative: bool = data.get('collaborative')
        self.description: str = data.get('description')
        self.followers: int = data.get('followers', {}).get('total', 0)

        self.tracks: List[PlaylistTrack] = [
            PlaylistTrack(track) for track in data['tracks']['items']
        ] if 'items' in data['tracks'] else []

        self.images: List[Image] = [
            Image(**image_) for image_ in data.get('images', [])
        ]

        self.total_tracks: int = data.get('tracks')['total']

    def __repr__(self) -> str:
        return f"melo.Playlist - {(self.name or self.id or self.uri)!r}"

    def get_tracks(
        self,
        limit: Optional[int] = 20,
        offset: Optional[int] = 0
    ) -> List[PlaylistTrack]:
        '''Get  a list of tracks based on the given limit and offset

        Parameters
        ----------
        limit: `int`
            The maximum number of tracks to return.
        offset: `int`
            Specifies where the API should begin fetching tracks from.

        Returns
        -------
        result: `List[PlaylistTrack]`
            A list of tracks from the playlist.
        '''

        if (offset + limit) < self.total_tracks:
            return [PlaylistTrack(track_) for track_ in self.tracks[offset: offset + limit]]

        # TODO - Use multi-type results with playlist items (it returns episodes as well)
        data = SPOTIFY.playlist_tracks(
            self.id,
            limit=limit,
            offset=offset
        )

        return [PlaylistTrack(track_) for track_ in data['items']]

    def get_all_tracks(self) -> List[PlaylistTrack]:
        '''Get a list of all the tracks in aplaylist

        This operation might take long depending on the size of the playlist.
        '''

        if len(self.tracks) >= self.total_tracks:
            return self.tracks

        self.tracks = []
        offset = 0

        while len(self.tracks) < self.total_tracks:
            data = SPOTIFY.playlist_tracks(
                self.id,
                limit=50,
                offset=offset
            )

            self.tracks.extend([PlaylistTrack(track_) for track_ in data])
            offset += 50

        self.total_tracks = len(self.tracks)
        return self.tracks

    def add_items(self, items: List[Union[Track, "spotify.Episode", str]]) -> None:
        from .episode import Episode

        items = list(
            map(
                lambda item: item.id if isinstance(item, (Track, Episode))
                else item,
                items
            )
        )
        SPOTIFY.playlist_add_items(self.id, items=items)
        if len(self.tracks) == self.total_tracks:
            self.tracks.extend(items)
        self.total_tracks += len(items)

    # def saved(self) -> bool:
    #     from .client import client

    #     return SPOTIFY.playlist_is_following(self.id, client.id)

    # def save_playlist(self) -> None:
    #     '''Add the current user as the follower of this playlist'''
    #     SPOTIFY.current_user_follow_playlist(self.id)

    # def unsave_playlist(self) -> None:
    #     return SPOTIFY.current_user_unfollow_playlist(self.id)

    # TODO - implement scopes and playlist modifications



import threading
import time

from ffpyplayer.player import MediaPlayer

FADE_IN_STEPS = [x / 10000000.0 for x in range(00000, 10078125, 78125)]
FADE_OUT_STEPS = [x / 10000000.0 for x in range(00000, 10078125, 78125)]
FADE_OUT_STEPS.reverse()
SMOOTHNESS_FACTOR = 32
FF_OPTS = {
    'paused': True,
    'vn': True,
    'sync': 'audio',
    'genpts': True,
    'infbuf': True
}


def close_player(player):
    # print(f':: CLOSING PLAYER at {player.get_pts()}')
    player.toggle_pause()
    player.close_player()
    time.sleep(1)
    # print(':: BROKE')


def close_stream(player: MediaPlayer, fade: bool = True) -> None:
    # #print(f'::: current_pts: {current_pts}, {DURATION - current_pts} seconds left')

    # print('::: entered closing')
    if fade:
        player_fade_out(player)
    else:
        close_player(player)


def player_fade_in(player: MediaPlayer, fade: int = 0):
    for step in FADE_IN_STEPS:
        delta = fade - player.get_pts()
        player.set_volume(step)
        #print(player.get_volume(), player.get_pts(), delta, step)
        time.sleep(delta / SMOOTHNESS_FACTOR)


def player_fade_out(player: MediaPlayer):
    DURATION = player.get_metadata()['duration']

    # print('::: entered closing- starting fade out')
    for step in FADE_OUT_STEPS[FADE_OUT_STEPS.index(player.get_volume()):]:
        delta = DURATION - player.get_pts()
        player.set_volume(step)
        #print(player.get_volume(), player.get_pts(), delta, step)
        time.sleep(delta / SMOOTHNESS_FACTOR)
    else:
        close_player(player)


def manage_stream(player: MediaPlayer, source: str, fade: int = 0, fade_in: bool = True) -> None:
    print('started thread')

    last_pts = 10
    updated_pts = 0

    if player.get_pause():
        player.toggle_pause()
    print('lmao')

    threading.Thread(
        target=player_fade_in,
        args=(player, fade)
    ).start() if fade_in else player.set_volume(1.0)

    DURATION = player.get_metadata()['duration']

    last_buffered_pts = 0
    buffer_repeat_count = 0
    #print(f":: METADATA: {player.get_metadata()}")
    while True:

        # print(updated_pts)
        updated_pts = int(str(player.get_pts()).split('.', maxsplit=1)[0])

        while player.get_pause():
            time.sleep(0.4)

        if (updated_pts == last_pts != 0 and
                not updated_pts == last_pts == DURATION):
            player.toggle_pause()
            #print(f"buffered out, pausing: {buffer_repeat_count}")
            # player.seek(-1, relative=True, accurate=False)
            player.toggle_pause()
            time.sleep(2)
            updated_pts = int(str(player.get_pts()).split('.', maxsplit=1)[0])
            last_buffered_pts = last_pts
            if last_buffered_pts == updated_pts:
                buffer_repeat_count += 1

        if updated_pts != last_buffered_pts:
            last_buffered_pts = 0
            buffer_repeat_count = 0

        if buffer_repeat_count == 2:
            #print('reviving stream')
            player.toggle_pause()
            player.close_player()
            time.sleep(1)
            del player
            #print('::: closed stream')
            time.sleep(1)
            player = MediaPlayer(source, ff_opts=FF_OPTS)
            time.sleep(2)
            #print('::: created new player')
            # player.seek(updated_pts + 1, accurate=False)
            player.seek(updated_pts, accurate=False)
            time.sleep(1)
            player.toggle_pause()
            buffer_repeat_count = 0

        # fade aka refenrece point to begin closing
        current_pts = int(str(player.get_pts()).split('.', maxsplit=1)[0])
        if current_pts + abs(fade) >= DURATION:
            if fade <= 3:
                if fade > 0:
                    time.sleep(fade)
                buffer_repeat_count = 0
                close_player(player)
                break
            # print(f"breaking from the handler thread, at timestamp: {current_pts}")
            threading.Thread(
                target=close_stream,
                args=(player, bool(fade))
            ).start()
            break
        last_pts = updated_pts
        time.sleep(1)
    #print('\r::: broke..\n>>> ', end='')



import io
import os
import threading
import time
from typing import BinaryIO, List, Union

from ffpyplayer.player import MediaPlayer
from ffpyplayer.tools import loglevels, set_log_callback

import melodine
from melodine.configs import TEMPFILES_DIR
from melodine.models import base
from melodine.player.helpers import manage_stream
from melodine.utils import singleton

loglevel_emit = 'error'


def log_callback(message, level):
    if message and loglevels[level] <= loglevels[loglevel_emit]:
        print("error")


set_log_callback(log_callback)

FF_OPTS = {
    'paused': True,
    'sync': 'audio',
    'vn': True,
    'genpts': True,
    'infbuf': True
}


def playsound(location: Union[str, BinaryIO], *, blocking: bool = False):
    file_name = None
    if isinstance(location, io.IOBase):
        file_mode = 'w' + 'b' if isinstance(location, io.FileIO) else ''
        file_name = os.path.join(TEMPFILES_DIR, id(location))
        with open(file_name, file_mode) as file:
            file.write(location.read())
        location = file_name

    player = MediaPlayer(location, ff_opts=FF_OPTS)
    time.sleep(1)
    player.toggle_pause()
    # time.sleep(2)

    manager_thread = threading.Thread(
        target=manage_stream, args=(player, location))
    if blocking:
        manager_thread.run()
    else:
        manager_thread.start()
    if file_name is None:
        os.remove(file_name)


def play(location: str, *, blocking: bool = False, fade: int = 0, fade_in: bool = False) -> Union[MediaPlayer, None]:
    time.sleep(0.5)
    player = MediaPlayer(location, ff_opts={**FF_OPTS, **{'volume': 0.0}})
    time.sleep(1.5)
    player.set_volume(0.0)
    time.sleep(0.5)

    manager_thread = threading.Thread(
        target=manage_stream, args=(player, location, fade), daemon=True)
    if blocking:
        manager_thread.run()
    elif blocking is None:
        return player, manager_thread
    else:
        manager_thread.start()
        return player


@singleton
class Player:
    def __init__(self, autoplay: bool = False, crossfade: int = 6):
        self.crossfade: int = crossfade
        self.autoplay: bool = autoplay

        self.shuffle: bool = False
        self.repeast: bool = False

        self.recently_played: List = []
        self.now_playing = None
        self.__now_playing = None
        self.queue: List[base.TrackBase] = []

        thread = threading.Thread(
            target=self.player_handler, args=(), daemon=True)
        thread.start()

    def player_handler(self):
        sleep_time = 0
        while True:
            for track in self.queue:
                print(f'new track - {track}')
                self.queue.remove(track)
                self.now_playing = track
                self.__now_playing, self.__now_playing_thread = play(
                    self.now_playing.url,
                    blocking=None,
                    fade=self.crossfade,
                    fade_in=True if len(self.queue) > 0 else False
                )
                print(self.__now_playing)
                print(self.now_playing)
                self.__now_playing_thread.run()
                self.recently_played.append(track)
                print(self.queue)
            time.sleep(0.4)

    # pause state
    def pause(self):
        self.__now_playing.set_pause(True) if self.__now_playing else ''

    def get_state(self) -> bool:
        return self.__now_playing.get_pause() if self.__now_playing else False

    def set_state(self, state: bool) -> None:
        self.__now_playing.set_pause(state) if self.__now_playing else ''

    def toggle_state(self) -> None:
        self.__now_playing.toggle_pause() if self.__now_playing else ''

    # player volume
    def get_volume(self) -> int:
        return self.__now_playing.get_volume() if self.__now_playing else 0

    def set_volume(self, volume: int) -> None:
        self.__now_playing.set_volume(volume) if self.__now_playing else ''

    # mute
    def get_mute(self) -> bool:
        return self.__now_playing.get_mute() if self.__now_playing else False

    def set_mute(self, mute: bool) -> None:
        self.__now_playing.set_mute(mute) if self.__now_playing else ''

    def toggle_mute(self) -> None:
        self.__now_playing.set_mute(
            not self.__now_playing.get_mute()) if self.__now_playing else ''

    # playback progress
    def get_current_timestamp(self) -> int:
        return int(self.__now_playing.get_pts()) if self.__now_playing else 0

    def seek(self, pts: int) -> None:
        self.__now_playing.seek(pts, relative=False,
                                accurate=False) if self.__now_playing else ''

    def get_shuffle(self) -> bool:
        return self.shuffle

    def set_shuffle(self, shuffle: bool) -> None:
        self.shuffle = shuffle

    def toggle_shuffle(self) -> None:
        self.shuffle = not self.shuffle

    # queing
    def play_next(self, track):
        self.queue.insert(0, track)

    def add_to_queue(self, item: Union[base.TrackBase, base.PlaylistBase, base.AlbumBase]):
        # if isinstance(item, (base.PlaylistBase, base.AlbumBase)):
        #     for track in item.tracks:
        #         self.queue.append(track)
        # else:
        self.queue.append(item)

    def remove_from_queue(self, track: Union[int, base.TrackBase]) -> None:
        if isinstance(track, int):
            del self.queue[track]
        else:
            self.queue.remove(track)

    # playback
    def play(self, track: Union[Union[base.TrackBase, base.ArtistBase, base.PlaylistBase, base.AlbumBase], None] = None) -> None:
        if track is None:
            self.__now_playing.set_pause(False)
        else:
            self.add_to_queue(track)

    def next(self): ...
    def previous(self): ...
    def rewind(self): ...

    # for duration / stream metadata (like audio quality)
    def get_metadata(self):
        return self.__now_playing.get_metadata() if self.__now_playing else {}



from typing import Any, Dict, List

from melodine.configs import YTMUSIC
from melodine.innertube import InnerTube
from melodine.models import ytmusic
from melodine.models.base.track import TrackBase
from melodine.utils import Image, URIBase


class Track(URIBase):
    '''A YTMusic Track Object

    Attributes
    ---------
    id: str
        The Spotify ID for the Track
    '''

    __slots__ = [
        'id',
        'name',
        'href',
        'uri',
        'explicit',
        'duration',
        'images',
        '_artists',
        '_album',

        '_url'
        '_recs',
        '_recs_offset'
    ]

    def __init__(self, data: Dict, **kwargs) -> None:
        self.id: str = data.get('videoId')
        self.name: str = data.get('title')
        self.href: str = "https://music.youtube.com/watch?v=" + self.id
        self.uri: str = f"ytmusic:track:{self.id}"

        self.explicit: bool = data.get('isExplicit', False)
        self.duration: int = data.get(
            'duration_seconds', data.get('lengthSeconds', 0))

        images: List[Dict]
        if 'thumbnail' in data:
            images = data.get('thumbnail')
            # sometimes, the images are in yet another nested dict
            if 'thumbnails' in images:
                # because it's a dict
                images = images.get('thumbnails')
        else:
            images = data.get('thumbnails')
        self.images: List[Image] = [
            Image(**image) for image in (images or [])
        ]

        artists: Any
        if 'artists' in data:
            artists = data.get('artists')
        else:
            artists = [{
                'name': data.get('author'),
                'id': data.get('channelId')
            }]
        self._artists: List[Dict] = artists

        album: Any
        if 'album' in data and not isinstance(data['album'], str):
            album = data.get('album')
        elif 'album' in kwargs:
            album = kwargs.get('album')
        else:
            album = None
        self._album: Dict = album

        self._url = str()
        self._recs = list()
        self._recs_offset = int()  # aka 0

    @classmethod
    def from_id(cls, id: str) -> "Track":
        '''tracks obtained directly from an ID have the album attribute as None'''
        return cls(data=YTMUSIC.get_song(id)['videoDetails'])

    @property
    def artists(self) -> List["ytmusic.Artist"]:
        """Get a list of all Artists from the track"""
        from .artist import Artist

        for idx, artist in enumerate(self._artists.copy()):
            if not isinstance(artist, Artist):
                artist = Artist.partial(artist)
                self._artists.insert(idx, artist)
                del self._artists[idx + 1]
        return self._artists

    @property
    def album(self) -> "ytmusic.Album":
        '''Get the Album which the track belongs to'''
        from .album import Album

        if self._album is not None and not isinstance(self._album, Album):
            self._album = Album.partial(self._album)
        return self._album

    @property
    def url(self) -> str:
        '''fetch the playback url for the track.'''
        if not self._url:
            video_id = YTMUSIC.search(
                f"{self.artists[0].name} - {self.name}", filter='songs'
            )[0]['videoId'] if not self.id else self.id

            video_info = InnerTube().player(video_id)

            # print(video_info['streamingData']['expiresInSeconds'])
            # print(video_info['streamingData']['formats'][0].keys())
            self._url = video_info['streamingData']['formats'][-2]['url']

            return self._url
        return self._url

    def cache_url(self) -> None:
        '''just a dummy call to trigger the url fetching.'''
        if not self.url:
            self._url = self.url

    @property
    def recommendations(self) -> List["Track"]:
        '''A list of recommended tracks for this track already fetched upto this point.

        it is initially empty and gets populated as more calls are made to `get_recommendations` are made.

        use `get_recommendations` to fetch new recommendations. instead'''
        return self._recs

    def get_recommendations(self, limit: int = 10) -> List["Track"]:
        '''
        used to get new recommendations for the track.

        earlier fetched recommended tracks can be accessed whith `.recommendations
        '''

        data = YTMUSIC.get_watch_playlist(
            self.id,
            f"RDAMVM{self.id}",
            limit=self._recs_offset + limit
        )

        tracks = data['tracks'][self._recs_offset: (self._recs_offset + limit)]
        tracks = [Track(track) for track in tracks]
        self._recs.extend(tracks)
        self._recs_offset += limit

        return tracks
